import numpy as np

from buzzard import _tools
from buzzard._footprint import Footprint
from buzzard._a_stored import AStored, ABackStored
from buzzard._a_source_raster import ASourceRaster, ABackSourceRaster

class AStoredRaster(AStored, ASourceRaster):
    """Base abstract class defining the common behavior of all rasters that are stored somewhere
    (like RAM or disk).

    Features Defined
    ----------------
    - Has a `set_data` method that allows to write pixels to storage
    """

    def set_data(self, array, fp=None, channels=None, interpolation='cv_area', mask=None, **kwargs):
        """.. _raster file set_data:

        Write a rectangle of data to the destination raster. Each channel in `array` is written to
        one channel in `raster` in the same order as described by the `channels` parameter. An
        optional `mask` may be provided to only write certain pixels of `array`.

        If `fp` is not fully within the destination raster, only the overlapping pixels are
        written.
        If `fp` is not on the same grid as the destination raster, remapping is automatically
        performed using the `interpolation` algorithm. (It fails if the `allow_interpolation`
        parameter is set to `False` in `Dataset` (default)). When interpolating:

        - The nodata values are not interpolated, they are correctly spread to the output.
        - At most one pixel may be lost at edges due to interpolation. Provide more context in
          `array` to compensate this loss.
        - The mask parameter is also interpolated.

        The alpha bands are currently resampled like any other band, this behavior may change in
        the future.

        This method is not thread-safe.

        Parameters
        ----------
        array: numpy.ndarray of shape (Y, X) or (Y, X, C)
            The values to be written
        fp: Footprint of shape (Y, X) or None
            If None: write the full source raster
            If Footprint: write this window to the raster
        channels: None or int or slice or sequence of int (see `Channels Parameter` below)
            The channels to be written.
        interpolation: one of {'cv_area', 'cv_nearest', 'cv_linear', 'cv_cubic', 'cv_lanczos4'} or None
            OpenCV method used if intepolation is necessary
        mask: numpy array of shape (Y, X) and dtype `bool` OR inputs accepted by `Footprint.burn_polygons`
            ..

        Channels Parameter
        ------------------
        +------------+-----------------------------------------------------+----------------+
        | type       | value                                               | meaning        |
        +============+=====================================================+================+
        | NoneType   | None (default)                                      | All channels   |
        +------------+-----------------------------------------------------+----------------+
        | slice      | slice(None), slice(1), slice(0, 2), slice(2, 0, -1) | Those channels |
        +------------+-----------------------------------------------------+----------------+
        | int        | 0, 1, 2, -1, -2, -3                                 | Channel `idx`  |
        +------------+-----------------------------------------------------+----------------+
        | (int, ...) | [0], [1], [2], [-1], [-2], [-3], [0, 1], [-1, 2, 1] | Those channels |
        +------------+-----------------------------------------------------+----------------+

        Caveat
        ------
        When using a Raster backed by a driver (like a GDAL driver), the data might be flushed to
        disk only after the garbage collection of the driver object. To be absolutely sure that the
        driver cache is flushed to disk, call `.close` or `.deactivate` on this Raster.

        """

        if self.mode != 'w': # pragma: no cover
            raise RuntimeError('Cannot write a read-only raster file')

        def _band_to_channels(val):
            val = np.asarray(val)
            if np.array_equal(val, -1):
                return None
            if val.ndim == 0:
                return val - 1
            if val.ndim != 1:
                raise ValueError('Error in deprecated `band` parameter')
            val = [
                v
                for v in val
                for v in (range(len(self)) if v == -1 else [v - 1])
            ]
            return val
        channels, kwargs = _tools.deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='channels', old_names={'band': '0.6.0'}, context='ASourceRaster.set_data',
            new_name_value=channels,
            new_name_is_provided=channels is not None,
            user_kwargs=kwargs,
            transform_old=_band_to_channels,
        )
        if kwargs: # pragma: no cover
            raise TypeError("set_data() got an unexpected keyword argument '{}'".format(
                list(kwargs.keys())[0]
            ))

        # Normalize and check fp parameter
        if fp is None:
            fp = self.fp
        elif not isinstance(fp, Footprint):
            raise ValueError('`fp` parameter should be a Footprint (not {})'.format(fp)) # pragma: no cover

        # Normalize and check channels parameter
        channel_ids, _ = _tools.normalize_channels_parameter(channels, len(self))
        if len(channel_ids) != len(set(channel_ids)): # pragma: no cover
            raise ValueError("The `channels` parameter should not reference twice the same channel")
        del channels

        # Normalize and check array parameter
        array = np.atleast_3d(array)
        if array.ndim != 3: # pragma: no cover
            raise ValueError('Input array should have 2 or 3 dimensions, not {}'.format(array.ndim))
        if array.shape[:2] != tuple(fp.shape): # pragma: no cover
            msg = 'Incompatible shape between input `array` ({}) and `fp` ({})'.format(
                array.shape[:2], tuple(fp.shape)
            )
            raise ValueError(msg)
        if len(channel_ids) != array.shape[-1]: # pragma: no cover
            msg = 'Incompatible number of channels between `array` ({}) and `channels` ({})'.format(
                len(channel_ids), array.shape[-1]
            )
            raise ValueError(msg)

        # Normalize and check mask parameter
        if mask is not None:
            if isinstance(mask, np.ndarray):
                mask = mask.astype(bool, copy=False)
                if mask.ndim != 2: # pragma: no cover
                    raise ValueError('Input `mask` should 2 dimensions')
                if mask.shape[:2] != tuple(fp.shape): # pragma: no cover
                    raise ValueError('Incompatible shape between input `mask` and `fp`')
            else:
                mask = fp.burn_polygons(mask)

        # Check interpolation parameter here
        if not (interpolation is None or interpolation in self._back.REMAP_INTERPOLATIONS): # pragma: no cover
            raise ValueError('`interpolation` should be None or one of {}'.format(
                set(self._back.REMAP_INTERPOLATIONS.keys())
            ))

        return self._back.set_data(
            array=array,
            fp=fp,
            channel_ids=channel_ids,
            interpolation=interpolation,
            mask=mask,
        )

    def fill(self, value, channels=None, **kwargs):
        """Fill raster with value.

        This method is not thread-safe.

        Parameters
        ----------
        value: nbr
            ..
        channels: int or sequence of int (see `Channels Parameter` below)
            The channels to be written

        Channels Parameter
        ------------------
        +------------+-----------------------------------------------------+----------------+
        | type       | value                                               | meaning        |
        +============+=====================================================+================+
        | NoneType   | None (default)                                      | All channels   |
        +------------+-----------------------------------------------------+----------------+
        | slice      | slice(None), slice(1), slice(0, 2), slice(2, 0, -1) | Those channels |
        +------------+-----------------------------------------------------+----------------+
        | int        | 0, 1, 2, -1, -2, -3                                 | Channel `idx`  |
        +------------+-----------------------------------------------------+----------------+
        | (int, ...) | [0], [1], [2], [-1], [-2], [-3], [0, 1], [-1, 2, 1] | Those channels |
        +------------+-----------------------------------------------------+----------------+

        Caveat
        ------
        When using a Raster backed by a driver (like a GDAL driver), the data might be flushed to
        disk only after the garbage collection of the driver object. To be absolutely sure that the
        driver cache is flushed to disk, call `.close` or `.deactivate` on this Raster.

        """
        if self.mode != 'w': # pragma: no cover
            raise RuntimeError('Cannot write a read-only raster file')

        def _band_to_channels(val):
            val = np.asarray(val)
            if np.array_equal(val, -1):
                return None
            if val.ndim == 0:
                return val - 1
            if val.ndim != 1:
                raise ValueError('Error in deprecated `band` parameter')
            val = [
                v
                for v in val
                for v in (range(len(self)) if v == -1 else [v - 1])
            ]
            return val
        channels, kwargs = _tools.deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='channels', old_names={'band': '0.6.0'}, context='ASourceRaster.fill',
            new_name_value=channels,
            new_name_is_provided=channels is not None,
            user_kwargs=kwargs,
            transform_old=_band_to_channels,
        )
        if kwargs: # pragma: no cover
            raise TypeError("fill() got an unexpected keyword argument '{}'".format(
                list(kwargs.keys())[0]
            ))

        channel_ids, _ = _tools.normalize_channels_parameter(channels, len(self))
        del channels
        channel_ids = set(channel_ids)

        value = self.dtype.type(value).tolist()

        self._back.fill(
            value=value,
            channel_ids=channel_ids,
        )

class ABackStoredRaster(ABackStored, ABackSourceRaster):
    """Implementation of AStoredRaster's specifications"""

    def set_data(self, array, fp, channels, interpolation, mask, **kwargs): # pragma: no cover
        raise NotImplementedError('ABackStoredRaster.set_data is virtual pure')

    def fill(self, value, channels, **kwargs): # pragma: no cover
        raise NotImplementedError('ABackStoredRaster.fill is virtual pure')



"""
TODO: Test insert_data under several modes
"""

# pylint: disable=redefined-outer-name, unused-argument

from __future__ import division, print_function
import os
import tempfile
import uuid
import string

import numpy as np
from osgeo import gdal
import pytest
import shapely.ops

import buzzard as buzz
from buzzard.test.tools import fpeq, sreq, eq
from buzzard.test import make_tile_set
from .tools import  get_srs_by_name

SR1 = get_srs_by_name('EPSG:2154')
SR2 = get_srs_by_name('EPSG:2154 centered')

# FIXTURES ************************************************************************************** **
@pytest.fixture(scope='module')
def fps():
    """See make_tile_set
    A B C
    D E F
    G H I
    """
    return make_tile_set.make_tile_set(3, [0.1, -0.1])

@pytest.fixture()
def env():
    with buzz.Env(significant=8):
        yield

@pytest.fixture(scope='module')
def shp1_path(fps):
    """Create a shapefile in SR1 containing all single letter polygons from `fps` fixture"""
    path = '{}/{}.shp'.format(tempfile.gettempdir(), uuid.uuid4())

    ds = buzz.Dataset()
    ds.create_vector('poly', path, 'polygon', sr=SR1['wkt'])
    for letter in string.ascii_uppercase[:9]:
        ds.poly.insert_data(fps[letter].poly)
    ds.poly.close()
    del ds
    yield path
    gdal.GetDriverByName('ESRI Shapefile').Delete(path)

@pytest.fixture(scope='module')
def tif1_path(fps):
    """Create a tif in SR1 with all single letter footprints from `fps` fixture burnt in it"""
    path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())

    ds = buzz.Dataset()
    ds.create_raster('rast', path, fps.AI, 'int32', 1, sr=SR1['wkt'])
    for letter in string.ascii_uppercase[:9]:
        fp = fps[letter]
        arr = np.full(fp.shape, ord(letter), dtype=int)
        ds.rast.set_data(arr, fp=fp)
    ds.rast.close()
    del ds
    yield path
    gdal.GetDriverByName('GTiff').Delete(path)

@pytest.fixture(scope='module')
def shp2_path(fps):
    """Create a shapefile in SR2 containing all single letter polygons from `fps` fixture"""
    path = '{}/{}.shp'.format(tempfile.gettempdir(), uuid.uuid4())

    ds = buzz.Dataset()
    ds.create_vector('poly', path, 'polygon', sr=SR2['wkt'])
    for letter in string.ascii_uppercase[:9]:
        ds.poly.insert_data(fps[letter].poly)
    ds.poly.close()
    del ds
    yield path
    gdal.GetDriverByName('ESRI Shapefile').Delete(path)

@pytest.fixture(scope='module')
def tif2_path(fps):
    """Create a tif in SR2 with all single letter footprints from `fps` fixture burnt in it"""
    path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())

    ds = buzz.Dataset()
    with ds.acreate_raster(path, fps.AI, 'int32', 1, sr=SR2['wkt']).close as r:
        for letter in string.ascii_uppercase[:9]:
            fp = fps[letter]
            arr = np.full(fp.shape, ord(letter), dtype=int)
            r.set_data(arr, fp=fp)
    yield path
    gdal.GetDriverByName('GTiff').Delete(path)


@pytest.fixture(scope='module')
def shp3_path(fps):
    """Create a shapefile without SR containing all single letter polygons from `fps` fixture"""
    path = '{}/{}.shp'.format(tempfile.gettempdir(), uuid.uuid4())

    ds = buzz.Dataset()
    ds.create_vector('poly', path, 'polygon', sr=None)
    for letter in string.ascii_uppercase[:9]:
        ds.poly.insert_data(fps[letter].poly)
    ds.poly.close()
    del ds
    yield path
    gdal.GetDriverByName('ESRI Shapefile').Delete(path)

@pytest.fixture(scope='module')
def tif3_path(fps):
    """Create a tif witwhout SR with all single letter footprints from `fps` fixture burnt in it"""
    path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())

    ds = buzz.Dataset()
    with ds.acreate_raster(path, fps.AI, 'int32', 1, sr=None).close as r:
        for letter in string.ascii_uppercase[:9]:
            fp = fps[letter]
            arr = np.full(fp.shape, ord(letter), dtype=int)
            r.set_data(arr, fp=fp)
    yield path
    gdal.GetDriverByName('GTiff').Delete(path)

@pytest.fixture()
def random_path_shp():
    """Create a temporary path, and take care of cleanup afterward"""
    path = '{}/{}.shp'.format(tempfile.gettempdir(), uuid.uuid4())
    yield path
    if os.path.isfile(path):
        try:
            gdal.GetDriverByName('ESRI Shapefile').Delete(path)
        except:
            os.remove(path)

@pytest.fixture()
def random_path_tif():
    """Create a temporary path, and take care of cleanup afterward"""
    path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())
    yield path
    if os.path.isfile(path):
        try:
            gdal.GetDriverByName('GTiff').Delete(path)
        except:
            os.remove(path)

# TESTS ***************************************************************************************** **
def test_mode1(fps, shp1_path, tif1_path, shp2_path, tif2_path, shp3_path, tif3_path):
    ds = buzz.Dataset()
    ds.open_raster('tif1', tif1_path)
    ds.open_vector('shp1', shp1_path)
    ds.open_raster('tif2', tif2_path)
    ds.open_vector('shp2', shp2_path)
    ds.open_raster('tif3', tif3_path)
    ds.open_vector('shp3', shp3_path)

    # Test SR equality
    assert sreq(
        ds.tif1.wkt_virtual,
        ds.tif1.wkt_stored,
        ds.tif1.proj4_virtual,
        ds.tif1.proj4_stored,
        ds.shp1.wkt_virtual,
        ds.shp1.wkt_stored,
        ds.shp1.proj4_virtual,
        ds.shp1.proj4_stored,
    )
    assert sreq(
        ds.tif2.wkt_virtual,
        ds.tif2.wkt_stored,
        ds.tif2.proj4_virtual,
        ds.tif2.proj4_stored,
        ds.shp2.wkt_virtual,
        ds.shp2.wkt_stored,
        ds.shp2.proj4_virtual,
        ds.shp2.proj4_stored,
    )
    assert not sreq(ds.tif1.wkt_stored, ds.tif2.wkt_stored)
    assert (
        None ==
        ds.wkt ==
        ds.proj4 ==
        ds.tif3.wkt_virtual ==
        ds.tif3.wkt_stored ==
        ds.tif3.proj4_virtual ==
        ds.tif3.proj4_stored ==
        ds.shp3.wkt_virtual ==
        ds.shp3.wkt_stored ==
        ds.shp3.proj4_virtual ==
        ds.shp3.proj4_stored
    )

    # Test footprints equality
    assert fpeq(
        fps.AI,
        # tif/shp 1
        ds.tif1.fp,
        ds.tif1.fp_origin,
        buzz.Footprint.of_extent(ds.shp1.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds[[0, 2, 1, 3]], fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),
        # tif/shp 2
        ds.tif2.fp,
        ds.tif2.fp_origin,
        buzz.Footprint.of_extent(ds.shp2.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds[[0, 2, 1, 3]], fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),
        # tif/shp 3
        ds.tif3.fp,
        ds.tif3.fp_origin,
        buzz.Footprint.of_extent(ds.shp3.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.bounds[[0, 2, 1, 3]], fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),
    )

    # Test what's written in all 6 files
    tif1 = ds.tif1.get_data()
    tif2 = ds.tif2.get_data()
    tif3 = ds.tif3.get_data()
    assert np.all(tif1 == tif2)
    assert np.all(tif1 == tif3)
    for i, letter in enumerate(string.ascii_uppercase[:9]):
        # tif/shp 1
        shp1 = ds.shp1.get_data(i, None)
        raster1_polys = ds.tif1.fp.find_polygons(tif1 == ord(letter))
        assert len(raster1_polys) == 1
        assert (shp1 ^ raster1_polys[0]).is_empty

        # tif/shp 2
        shp2 = ds.shp2.get_data(i, None)
        raster2_polys = ds.tif2.fp.find_polygons(tif2 == ord(letter))
        assert len(raster2_polys) == 1
        assert (shp2 ^ raster2_polys[0]).is_empty

        # tif/shp 3
        shp3 = ds.shp3.get_data(i, None)
        raster3_polys = ds.tif3.fp.find_polygons(tif3 == ord(letter))
        assert len(raster3_polys) == 1
        assert (shp3 ^ raster3_polys[0]).is_empty

def test_mode2(fps, shp1_path, tif1_path, shp2_path, tif2_path, shp3_path, tif3_path,
               random_path_shp, random_path_tif, env):
    ds = buzz.Dataset(sr_work=SR1['wkt'])
    ds.open_raster('tif1', tif1_path)
    ds.open_vector('shp1', shp1_path)
    ds.open_raster('tif2', tif2_path)
    ds.open_vector('shp2', shp2_path)

    # Test file creation/opening without spatial reference
    with buzz.Env(allow_complex_footprint=True):
        with pytest.raises(ValueError, match='spatial refe'):
            ds.acreate_vector(random_path_shp, 'polygon', [], sr=None)
        with pytest.raises(ValueError, match='spatial refe'):
            ds.acreate_raster(random_path_tif, fps.AI, 'int32', 1, {}, sr=None)
        with pytest.raises(ValueError, match='spatial refe'):
            ds.aopen_raster(tif3_path)
        with pytest.raises(ValueError, match='spatial refe'):
            ds.aopen_vector(shp3_path)

    # Test SR equality
    assert sreq(
        ds.wkt,
        ds.proj4,
        ds.tif1.wkt_virtual,
        ds.tif1.wkt_stored,
        ds.tif1.proj4_virtual,
        ds.tif1.proj4_stored,
        ds.shp1.wkt_virtual,
        ds.shp1.wkt_stored,
        ds.shp1.proj4_virtual,
        ds.shp1.proj4_stored,
    )
    assert sreq(
        ds.tif2.wkt_virtual,
        ds.tif2.proj4_virtual,
        ds.shp2.wkt_virtual,
        ds.shp2.proj4_virtual,
        ds.tif2.wkt_stored,
        ds.tif2.proj4_stored,
        ds.shp2.wkt_stored,
        ds.shp2.proj4_stored,
    )
    assert not sreq(ds.tif1.wkt_stored, ds.tif2.wkt_stored)

    # Test foorprints equality
    assert fpeq(
        fps.AI,
        # tif/shp 1
        ds.tif1.fp,
        ds.tif1.fp_origin,
        buzz.Footprint.of_extent(ds.shp1.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds[[0, 2, 1, 3]], fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 2
        ds.tif2.fp_origin,
        buzz.Footprint.of_extent(ds.shp2.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),
    )
    assert fpeq(
        ds.tif2.fp,
        buzz.Footprint.of_extent(ds.shp2.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds[[0, 2, 1, 3]], fps.AI.scale),
    )
    assert ds.tif2.fp != ds.tif1.fp

    # Test file creation with/without conversion of footprint
    with buzz.Env(allow_complex_footprint=True):
        with ds.acreate_raster(random_path_tif, fps.AI, 'int32', 1, sr=SR1['wkt'], ow=1).delete as r:
            assert fpeq(
                fps.AI,
                r.fp,
                r.fp_origin
            )
        with ds.acreate_raster(random_path_tif, fps.AI, 'int32', 1, sr=SR2['wkt'], ow=1).delete as r:
            assert fpeq(
                fps.AI,
                r.fp,
            )
            assert fps.AI != r.fp_origin

    # Test what's written in all 4 files
    tif1 = ds.tif1.get_data()
    tif2 = ds.tif2.get_data()
    assert np.all(tif1 == tif2)

    def f(x, y, z=None):
        return np.around(x, 6), np.around(y, 6)

    for i, letter in enumerate(string.ascii_uppercase[:9]):
        # tif/shp 1
        shp1 = ds.shp1.get_data(i, None)
        shp1 = shapely.ops.transform(f, shp1)

        raster1_polys = ds.tif1.fp.find_polygons(tif1 == ord(letter))
        assert len(raster1_polys) == 1
        raster1_poly = raster1_polys[0]
        raster1_poly = shapely.ops.transform(f, shp1)

        assert (shp1 ^ raster1_poly).is_empty

        # tif/shp 2
        shp2 = ds.shp2.get_data(i, None)
        shp2 = shapely.ops.transform(f, shp2)

        raster2_polys = ds.tif2.fp.find_polygons(tif2 == ord(letter))
        assert len(raster2_polys) == 1
        raster2_poly = raster2_polys[0]
        raster2_poly = shapely.ops.transform(f, shp2)

        assert (shp2 ^ raster2_poly).is_empty

def test_mode3(fps, shp1_path, tif1_path, shp2_path, tif2_path, shp3_path, tif3_path,
               random_path_shp, random_path_tif, env):
    ds = buzz.Dataset(sr_work=SR1['wkt'], sr_fallback=SR2['wkt'])
    ds.open_raster('tif1', tif1_path)
    ds.open_vector('shp1', shp1_path)
    ds.open_raster('tif2', tif2_path)
    ds.open_vector('shp2', shp2_path)
    ds.open_raster('tif3', tif3_path)
    ds.open_vector('shp3', shp3_path)

    # Test file creation without spatial reference
    with buzz.Env(allow_complex_footprint=True):
        with ds.acreate_vector(random_path_shp, 'polygon', [], sr=None).close as r:
            assert r.wkt_stored == None
            assert sreq(r.wkt_virtual, SR2['wkt'])
        with ds.acreate_raster(random_path_tif, fps.AI, 'int32', 1, {}, sr=None).close as v:
            assert v.wkt_stored == None
            assert sreq(v.wkt_virtual, SR2['wkt'])

    # Test SR equality
    assert sreq(
        ds.wkt,
        ds.proj4,
        ds.tif1.wkt_virtual,
        ds.tif1.wkt_stored,
        ds.tif1.proj4_virtual,
        ds.tif1.proj4_stored,
        ds.shp1.wkt_virtual,
        ds.shp1.wkt_stored,
        ds.shp1.proj4_virtual,
        ds.shp1.proj4_stored,

    )
    assert sreq(
        ds.tif2.wkt_virtual,
        ds.tif2.proj4_virtual,
        ds.shp2.wkt_virtual,
        ds.shp2.proj4_virtual,
        ds.tif2.wkt_stored,
        ds.tif2.proj4_stored,
        ds.shp2.wkt_stored,
        ds.shp2.proj4_stored,

        ds.tif3.wkt_virtual,
        ds.tif3.proj4_virtual,
        ds.shp3.wkt_virtual,
        ds.shp3.proj4_virtual,
    )
    assert not sreq(ds.tif1.wkt_stored, ds.tif2.wkt_stored)
    assert (
        None ==
        ds.tif3.wkt_stored ==
        ds.tif3.proj4_stored ==
        ds.shp3.wkt_stored ==
        ds.shp3.proj4_stored
    )

    # Test foorprints equality
    assert fpeq(
        fps.AI,
        # tif/shp 1
        ds.tif1.fp,
        ds.tif1.fp_origin,
        buzz.Footprint.of_extent(ds.shp1.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds[[0, 2, 1, 3]], fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 2
        ds.tif2.fp_origin,
        buzz.Footprint.of_extent(ds.shp2.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 3
        ds.tif3.fp_origin,
        buzz.Footprint.of_extent(ds.shp3.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),
    )
    assert fpeq(
        ds.tif2.fp,
        buzz.Footprint.of_extent(ds.shp2.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds[[0, 2, 1, 3]], fps.AI.scale),

        ds.tif3.fp,
        buzz.Footprint.of_extent(ds.shp3.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.bounds[[0, 2, 1, 3]], fps.AI.scale),
    )
    assert ds.tif2.fp != ds.tif1.fp

    # Test what's written in all 6 files
    tif1 = ds.tif1.get_data()
    tif2 = ds.tif2.get_data()
    tif3 = ds.tif3.get_data()
    assert np.all(tif1 == tif2)
    assert np.all(tif1 == tif3)

    def f(x, y, z=None):
        return np.around(x, 6), np.around(y, 6)

    for i, letter in enumerate(string.ascii_uppercase[:9]):
        # tif/shp 1
        shp1 = ds.shp1.get_data(i, None)
        shp1 = shapely.ops.transform(f, shp1)

        raster1_polys = ds.tif1.fp.find_polygons(tif1 == ord(letter))
        assert len(raster1_polys) == 1
        raster1_poly = raster1_polys[0]
        raster1_poly = shapely.ops.transform(f, shp1)

        assert (shp1 ^ raster1_poly).is_empty

        # tif/shp 2
        shp2 = ds.shp2.get_data(i, None)
        shp2 = shapely.ops.transform(f, shp2)

        raster2_polys = ds.tif2.fp.find_polygons(tif2 == ord(letter))
        assert len(raster2_polys) == 1
        raster2_poly = raster2_polys[0]
        raster2_poly = shapely.ops.transform(f, shp2)

        assert (shp2 ^ raster2_poly).is_empty

        # tif/shp 3
        shp3 = ds.shp3.get_data(i, None)
        shp3 = shapely.ops.transform(f, shp3)

        raster3_polys = ds.tif3.fp.find_polygons(tif3 == ord(letter))
        assert len(raster3_polys) == 1
        raster3_poly = raster3_polys[0]
        raster3_poly = shapely.ops.transform(f, shp3)

        assert (shp3 ^ raster3_poly).is_empty

def test_mode4(fps, shp1_path, tif1_path, shp2_path, tif2_path, shp3_path, tif3_path,
               random_path_shp, random_path_tif, env):
    ds = buzz.Dataset(sr_work=SR1['wkt'], sr_forced=SR2['wkt'])
    ds.open_raster('tif1', tif1_path)
    ds.open_vector('shp1', shp1_path)
    ds.open_raster('tif2', tif2_path)
    ds.open_vector('shp2', shp2_path)
    ds.open_raster('tif3', tif3_path)
    ds.open_vector('shp3', shp3_path)

    # Test file creation without spatial reference
    with buzz.Env(allow_complex_footprint=True):
        with ds.acreate_vector(random_path_shp, 'polygon', [], sr=None).close as r:
            assert r.wkt_stored == None
            assert sreq(r.wkt_virtual, SR2['wkt'])
        with ds.acreate_raster(random_path_tif, fps.AI, 'int32', 1, {}, sr=None).close as v:
            assert v.wkt_stored == None
            assert sreq(v.wkt_virtual, SR2['wkt'])

    # Test SR equality
    assert sreq(
        ds.wkt,
        ds.proj4,
        ds.tif1.wkt_stored,
        ds.tif1.proj4_stored,
        ds.shp1.wkt_stored,
        ds.shp1.proj4_stored,
    )
    assert sreq(
        ds.tif1.wkt_virtual,
        ds.tif1.proj4_virtual,
        ds.shp1.wkt_virtual,
        ds.shp1.proj4_virtual,

        ds.tif2.wkt_virtual,
        ds.tif2.proj4_virtual,
        ds.shp2.wkt_virtual,
        ds.shp2.proj4_virtual,
        ds.tif2.wkt_stored,
        ds.tif2.proj4_stored,
        ds.shp2.wkt_stored,
        ds.shp2.proj4_stored,

        ds.tif3.wkt_virtual,
        ds.tif3.proj4_virtual,
        ds.shp3.wkt_virtual,
        ds.shp3.proj4_virtual,
    )
    assert not sreq(ds.tif1.wkt_stored, ds.tif2.wkt_stored)
    assert (
        None ==
        ds.tif3.wkt_stored ==
        ds.tif3.proj4_stored ==
        ds.shp3.wkt_stored ==
        ds.shp3.proj4_stored
    )

    # Test foorprints equality
    assert fpeq(
        fps.AI,
        # tif/shp 1
        ds.tif1.fp_origin,
        buzz.Footprint.of_extent(ds.shp1.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 2
        ds.tif2.fp_origin,
        buzz.Footprint.of_extent(ds.shp2.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 3
        ds.tif3.fp_origin,
        buzz.Footprint.of_extent(ds.shp3.extent_stored, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.bounds_stored[[0, 2, 1, 3]], fps.AI.scale),
    )
    assert fpeq(
        # tif/shp 1
        ds.tif1.fp,
        buzz.Footprint.of_extent(ds.shp1.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp1.bounds[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 2
        ds.tif2.fp,
        buzz.Footprint.of_extent(ds.shp2.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp2.bounds[[0, 2, 1, 3]], fps.AI.scale),

        # tif/shp 3
        ds.tif3.fp,
        buzz.Footprint.of_extent(ds.shp3.extent, fps.AI.scale),
        buzz.Footprint.of_extent(ds.shp3.bounds[[0, 2, 1, 3]], fps.AI.scale),
    )
    assert ds.tif1.fp != ds.tif1.fp_stored

    # Test what's written in all 6 files
    tif1 = ds.tif1.get_data()
    tif2 = ds.tif2.get_data()
    tif3 = ds.tif3.get_data()
    assert np.all(tif1 == tif2)
    assert np.all(tif1 == tif3)

    def f(x, y, z=None):
        return np.around(x, 6), np.around(y, 6)

    for i, letter in enumerate(string.ascii_uppercase[:9]):
        # tif/shp 1
        shp1 = ds.shp1.get_data(i, None)
        shp1 = shapely.ops.transform(f, shp1)

        raster1_polys = ds.tif1.fp.find_polygons(tif1 == ord(letter))
        assert len(raster1_polys) == 1
        raster1_poly = raster1_polys[0]
        raster1_poly = shapely.ops.transform(f, shp1)

        assert (shp1 ^ raster1_poly).is_empty

        # tif/shp 2
        shp2 = ds.shp2.get_data(i, None)
        shp2 = shapely.ops.transform(f, shp2)

        raster2_polys = ds.tif2.fp.find_polygons(tif2 == ord(letter))
        assert len(raster2_polys) == 1
        raster2_poly = raster2_polys[0]
        raster2_poly = shapely.ops.transform(f, shp2)

        assert (shp2 ^ raster2_poly).is_empty

        # tif/shp 3
        shp3 = ds.shp3.get_data(i, None)
        shp3 = shapely.ops.transform(f, shp3)

        raster3_polys = ds.tif3.fp.find_polygons(tif3 == ord(letter))
        assert len(raster3_polys) == 1
        raster3_poly = raster3_polys[0]
        raster3_poly = shapely.ops.transform(f, shp3)

        assert (shp3 ^ raster3_poly).is_empty



"""Private tools to normalize functions parameters"""

import collections
import logging
import functools
import numbers

import six
import numpy as np

from .helper_classes import Singleton
from . import conv

Footprint, AAsyncRaster = None, None # Lazy import

CHANNELS_SCHEMA_PARAMS = frozenset({
    'nodata', 'interpretation', 'offset', 'scale', 'mask',
})

def _coro_parameter_0or1dim(val, clean_fn, name):
    """Normalize a parameter that can be an sequence or not.

    Parameters
    ----------
    val: User input
    clean_fn: Function to clean a value
    name: Name given to the values for exceptions

    Yield values
    ------------
    First value: bool
        Is flat
    """

    attempt = clean_fn(val)
    if attempt is not None:
        yield True
        yield attempt
        return
    if isinstance(val, str) or not isinstance(val, collections.Iterable): # pragma: no cover
        raise TypeError(
            'Expecting a `{}` or an `sequence of `{}`, found a `{}`'.format(name, name, type(val))
        )
    yield False
    for elt in val:
        attempt = clean_fn(elt)
        if attempt is None: # pragma: no cover
            fmt = 'Expecting a `{}` or an `sequence of `{}`, found an `sequence of {}`'
            raise TypeError(fmt.format(name, name, type(elt)))
        yield attempt

def normalize_fields_parameter(fields, index_of_field_name):
    count = len(index_of_field_name)

    def _normalize_value(val):
        if np.all(np.isreal(val)) and np.shape(val) == ():
            val = int(val)
            if not (val == -1 or 0 <= val < count): # pragma: no cover
                raise ValueError('field index should be -1 or between 0 and {}'.format(count - 1))
            return val
        elif isinstance(val, str): # pragma: no cover
            if val not in index_of_field_name:
                raise ValueError('Unknown field name')
            return val
        return None

    def _index_generator(clean_indices_generator):
        for index in clean_indices_generator:
            if isinstance(index, str):
                yield index_of_field_name[index]
            elif isinstance(index, int):
                if index == -1:
                    for i in range(len(index_of_field_name)):
                        yield i
                else:
                    yield index
            else: # pragma: no cover
                assert False, index

    if fields is None:
        return [], True
    if isinstance(fields, str):
        fields = [
            f
            for f in fields.replace(' ', ',').split(',')
            if f
        ]
    gen = _coro_parameter_0or1dim(fields, _normalize_value, 'field index')
    is_flat = next(gen)
    indices = list(_index_generator(gen))
    if len(indices) > 0:
        is_flat = False
    return indices, is_flat

def normalize_channels_parameter(channels, channel_count):
    if channels is None:
        if channel_count == 1:
            return [0], True
        else:
            return list(range(channel_count)), False
    elif not isinstance(channels, slice):
        channels = np.asarray(channels).tolist()

    indices = np.arange(channel_count)
    indices = indices[channels]
    indices = np.atleast_1d(indices)

    if isinstance(channels, slice):
        return indices.tolist(), False

    channels = np.asarray(channels)
    if not np.issubdtype(channels.dtype, np.number):
        raise TypeError('`channels` should be None or int or slice or list of int')
    if channels.ndim == 0:
        assert len(indices) == 1
        return indices.tolist(), True
    return indices.tolist(), False

def sanitize_channels_schema(channels_schema, channel_count):
    """Used on file/recipe creation"""
    ret = {}

    def _test_length(val, name):
        count = len(val)
        if count > channel_count: # pragma: no cover
            raise ValueError('Too many values provided for %s (%d instead of %d)' % (
                name, count, channel_count
            ))
        elif count < channel_count: # pragma: no cover
            raise ValueError('Not enough values provided for %s (%d instead of %d)' % (
                name, count, channel_count
            ))

    if channels_schema is None:
        return {}
    diff = set(channels_schema.keys()) - CHANNELS_SCHEMA_PARAMS
    if diff: # pragma: no cover
        raise ValueError('Unknown channels_schema keys `%s`' % diff)

    def _normalize_multi_layer(name, val, is_type, cleaner, default):
        if val is None:
            for _ in range(channel_count):
                yield default
        elif is_type(val):
            val = cleaner(val)
            for _ in range(channel_count):
                yield val
        else:
            _test_length(val, name)
            for elt in val:
                if elt is None:
                    yield default
                elif is_type(elt):
                    yield cleaner(elt)
                else: # pragma: no cover
                    raise ValueError('`{}` cannot use value `{}`'.format(name, elt))

    if 'nodata' in channels_schema:
        ret['nodata'] = list(_normalize_multi_layer(
            'nodata',
            channels_schema['nodata'],
            lambda x: np.all(np.isreal(x)) and np.shape(x) == (),
            lambda x: float(np.asarray(x).item()),
            None,
        ))

    if 'interpretation' in channels_schema:
        val = channels_schema['interpretation']
        if isinstance(val, str):
            ret['interpretation'] = [conv.gci_of_str(val)] * channel_count
        else:
            _test_length(val, 'nodata')
            ret['interpretation'] = [conv.gci_of_str(elt) for elt in val]
        ret['interpretation'] = [conv.str_of_gci(v) for v in ret['interpretation']]

    if 'offset' in channels_schema:
        ret['offset'] = list(_normalize_multi_layer(
            'offset',
            channels_schema['offset'],
            lambda x: np.all(np.isreal(x)) and np.shape(x) == (),
            lambda x: float(np.asarray(x).item()),
            0.,
        ))

    if 'scale' in channels_schema:
        ret['scale'] = list(_normalize_multi_layer(
            'scale',
            channels_schema['scale'],
            lambda x: np.all(np.isreal(x)) and np.shape(x) == (),
            lambda x: float(np.asarray(x).item()),
            1.,
        ))

    if 'mask' in channels_schema:
        val = channels_schema['mask']
        if isinstance(val, str):
            ret['mask'] = [conv.gmf_of_str(val)] * channel_count
        else:
            _test_length(val, 'mask')
            ret['mask'] = [conv.gmf_of_str(elt) for elt in val]
            shared_bit = conv.gmf_of_str('per_dataset')
            shared = [elt for elt in ret['mask'] if elt & shared_bit]
            if len(set(shared)) > 1: # pragma: no cover
                raise ValueError('per_dataset mask must be shared with same flags')
        ret['mask'] = [conv.str_of_gmf(v) for v in ret['mask']]


    ret = {
        k: tuple(v)
        for k, v in ret.items()
    }
    return ret


class _DeprecationPool(Singleton):
    """Singleton class designed to handle function parameter renaming"""

    def __init__(self):
        self._seen = set()

    class _MethodWrapper(object):
        """Descriptor object to manage deprecation"""
        def __init__(self, method, deprecation_version, seen):
            self._method = method
            self._deprecation_version = deprecation_version
            self._seen = seen
            self._old_name = None
            self._key = None

        def __get__(self, instance, owner):
            assert self._key is not None
            if self._key not in self._seen:
                self._seen.add(self._key)
                logging.warning('`{}` is deprecated since v{}, use `{}` instead'.format(
                    self._old_name, self._deprecation_version, self._method.__name__,
                ))
            return self._method.__get__(instance, owner)

        def __set_name__(self, owner, name):
            self._old_name = name
            self._key = (self._method, name)

    class _PropertyWrapper(object):
        """Descriptor object to manage deprecation"""
        def __init__(self, new_property_name, deprecation_version, seen):
            self._new_property_name = new_property_name
            self._deprecation_version = deprecation_version
            self._seen = seen
            self._old_name = None
            self._key = None

        def __get__(self, instance, owner):
            assert self._key is not None
            if self._key not in self._seen:
                self._seen.add(self._key)
                logging.warning('`{}` is deprecated since v{}, use `{}` instead'.format(
                    self._old_name, self._deprecation_version, self._new_property_name,
                ))
            return getattr(instance, self._new_property_name)

        def __set_name__(self, owner, name):
            self._old_name = name
            self._key = (owner, self._new_property_name, name)


    def wrap_method(self, method, deprecation_version):
        return self._MethodWrapper(method, deprecation_version, self._seen)

    def wrap_property(self, new_property, deprecation_version):
        return self._PropertyWrapper(new_property, deprecation_version, self._seen)

    def wrap_class(self, cls, old_name, deprecation_version):
        key = (cls, old_name)
        @functools.wraps(cls.__init__)
        def _f(*args, **kwargs):
            if key not in self._seen:
                self._seen.add(key)
                logging.warning('`{}` is deprecated since v{}, use `{}` instead'.format(
                    old_name, deprecation_version, cls.__name__,
                ))
            return cls.__init__(*args, **kwargs)
        return type(old_name, (cls,), {'__init__': _f})

    def handle_param_renaming_with_kwargs(self, new_name, old_names, context,
                                          new_name_value, new_name_is_provided, user_kwargs,
                                          transform_old=lambda x:x):
        """Look for errors with a particular parameter in an invocation

        Exemple
        -------
        >>> def fn(newname='default', **kwargs):
        ...     newname, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
        ...         new_name='newname', old_names={'oldname': '0.2.3'},
        ...         new_name_value=newname, context='the fn function',
        ...         new_name_is_provided=newname != 'default',
        ...         user_kwargs=kwargs,
        ...     )
        ...     return newname

        >>> fn() # Nothing happens
        'default'

        >>> fn(newname='default') # Nothing happens
        'default'

        >>> fn(oldname='aha') # A warning is issued the first time
        WARNING:root:`oldname` is deprecated since v0.2.3, use `newname`
        'aha'

        >>> fn(newname='default', oldname='default') # A warning is issued the first time
        WARNING:root:`oldname` is deprecated since v0.2.3, use `newname`
        'default'

        >>> fn(newname='default', oldname='aha') # A warning is issued the first time
        WARNING:root:`oldname` is deprecated since v0.2.3, use `newname`
        'aha'

        >>> fn(newname='aha', oldname='default') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        NameError: Using both `newname` and `oldname`, `oldname` is deprecated

        >>> fn(newname='aha', oldname='aha') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        NameError: Using both `newname` and `oldname`, `oldname` is deprecated

        """
        deprecated_names_used = six.viewkeys(old_names) & six.viewkeys(user_kwargs)
        if len(deprecated_names_used) == 0:
            return new_name_value, user_kwargs
        n = deprecated_names_used.pop()
        if new_name_is_provided:
            raise NameError('Using both `{}` and `{}` in `{}`, `{}` is deprecated'.format(
                new_name, n, context, n,
            ))

        key = (context, new_name, n)
        if key not in self._seen:
            self._seen.add(key)
            logging.warning(
                '`{}` parameter in `{}` is deprecated since v{}, use `{}` instead'.format(
                    n, context, old_names[n], new_name,
                )
            )
        v = transform_old(user_kwargs[n])
        del user_kwargs[n]
        return v, user_kwargs

    def handle_param_removal_with_kwargs(self, old_names, context, user_kwargs):
        deprecated_names_used = six.viewkeys(old_names) & six.viewkeys(user_kwargs)
        if len(deprecated_names_used) == 0:
            return user_kwargs
        n = deprecated_names_used.pop()
        key = (context, n)
        if key not in self._seen:
            self._seen.add(key)
            logging.warning('`{}` parameter in `{}` was removed in v{}'.format(
                n, context, old_names[n],
            ))
        del user_kwargs[n]
        return user_kwargs

deprecation_pool = _DeprecationPool()

def normalize_fields_defn(fields):
    """Used on file creation"""
    if not isinstance(fields, collections.Iterable): # pragma: no cover
        raise TypeError('Bad fields definition type')

    def _sanitize_dict(dic):
        dic = dict(dic)
        name = dic.pop('name')
        type_ = dic.pop('type')
        precision = dic.pop('precision', None)
        width = dic.pop('width', None)
        nullable = dic.pop('nullable', None)
        default = dic.pop('default', None)
        oft = conv.oft_of_any(type_)
        if default is not None:
            default = str(conv.type_of_oftstr(conv.str_of_oft(oft))(default))
        if len(dic) != 0: # pragma: no cover
            raise ValueError('unexpected keys in {} dict: {}'.format(name, dic))
        return dict(
            name=name,
            type=oft,
            precision=precision,
            width=width,
            nullable=nullable,
            default=default,
        )
    return [_sanitize_dict(dic) for dic in fields]

# Async rasters ***************************************************************************** **
def parse_queue_data_parameters(context, raster, channels=None, dst_nodata=None,
                                interpolation='cv_area', max_queue_size=5, **kwargs):
    """Check and transform the last parameters of a `queue_data` method.
    Default values are duplicated in the .queue_data and .iter_data methods
    """

    def _band_to_channels(val):
        val = np.asarray(val)
        if np.array_equal(val, -1):
            return None
        if val.ndim == 0:
            return val - 1
        if val.ndim != 1:
            raise ValueError('Error in deprecated `band` parameter')
        val = [
            v
            for v in val
            for v in (range(len(self)) if v == -1 else [v - 1])
        ]
        return val
    channels, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
        new_name='channels', old_names={'band': '0.6.0'}, context='Raster.{}'.format(context),
        new_name_value=channels,
        new_name_is_provided=channels is not None,
        user_kwargs=kwargs,
        transform_old=_band_to_channels,
    )
    if kwargs: # pragma: no cover
        raise TypeError("{}() got an unexpected keyword argument '{}'".format(
            context, list(kwargs.keys())[0]
        ))

    # Normalize and check channels parameter
    channel_ids, is_flat = normalize_channels_parameter(channels, len(raster))
    del channels

    # Normalize and check dst_nodata parameter
    if dst_nodata is not None:
        dst_nodata = raster.dtype.type(dst_nodata)
    elif raster.nodata is not None:
        dst_nodata = raster.nodata
    else:
        dst_nodata = raster.dtype.type(0)

    # Check interpolation parameter here
    if not (interpolation is None or interpolation in raster._back.REMAP_INTERPOLATIONS): # pragma: no cover
        raise ValueError('`interpolation` should be None or one of {}'.format(
            set(raster._back.REMAP_INTERPOLATIONS.keys())
        ))

    # Check max_queue_size
    max_queue_size = int(max_queue_size)
    if max_queue_size <= 0:
        raise ValueError('`max_queue_size` should be >0')

    return dict(
        channel_ids=channel_ids,
        dst_nodata=dst_nodata,
        interpolation=interpolation,
        max_queue_size=max_queue_size,
        is_flat=is_flat,
    )

def shatter_queue_data_method(met, name):
    """Check and transform a `met = queue_data_per_primitive[name]` given by user

    Parameters
    ----------
    met: object
        user's parameter
    name: hashable
        name given to primitive by user

    Returns
    -------
    (ABackAsyncRaster, dict of str->object)
    """
    global AAsyncRaster
    if AAsyncRaster is None:
        from buzzard._a_async_raster import AAsyncRaster


    # Unwrap function.partial instances ****************************************
    kwargs = {}
    while isinstance(met, functools.partial):
        if met.args:
            raise ValueError("Can't handle positional arguments in functools.partial " +
                             "of `queue_data_per_primitive` element")
        kwargs.update(met.keywords)
        met = met.func

    # Check method *************************************************************
    if not callable(met):
        raise TypeError('`queue_data_per_primitive[{}]` should be callable'.format(
            name
        ))
    if not hasattr(met, '__self__') or not isinstance(met.__self__, AAsyncRaster):
        fmt = '`queue_data_per_primitive[{}]` should be the `.queue_data` method ' +\
              'of a scheduler raster'
        raise TypeError(fmt.format(name))

    kwargs = parse_queue_data_parameters('create_raster_recipe', met.__self__, **kwargs)
    return met.__self__._back, kwargs

# Tiling checks ********************************************************************************* **
def is_tiling_covering_fp(tiling, fp, allow_outer_pixels, allow_overlapping_pixels):
    """Is the `tiling` object, an output of `fp.tile` or `fp.tile.count`?

    if `allow_outer_pixels`
        Some pixels of `tiling` may be outside of `fp`
    else
        All pixels of `tiling` should be inside `fp`

    if `allow_overlapping_pixels`
        All pixels of `fp` should be covered at least once.
    else
        All pixels of `fp` should be covered exactly once.

    Parameters
    ----------
    tiling: object
        An object provided by user
    fp: Footprint
    allow_outer_pixels: bool
    allow_overlapping_pixels: bool

    Returns
    -------
    bool
    """
    global Footprint
    if Footprint is None:
        from buzzard._footprint import Footprint

    # Type checking ****************************************
    if not isinstance(tiling, np.ndarray):
        return False
    if tiling.ndim != 2:
        return False
    for tile in tiling.flat:
        if not isinstance(tile, Footprint):
            return False
        if not fp.same_grid(tile):
            return False

    # Pixel indices extraction *****************************
    rtls = np.asarray([
        fp.spatial_to_raster(tile.tl)
        for tile in tiling.flat
    ]).reshape(tiling.shape[0], tiling.shape[1], 2)
    rbrs = np.asarray([
        fp.spatial_to_raster(tile.br)
        for tile in tiling.flat
    ]).reshape(tiling.shape[0], tiling.shape[1], 2)

    # is tiling ********************************************
    # All line's tly equal
    if not np.all(rtls[:, :1, 1] == rtls[:, :, 1]):
        return False
    # All line's bry equal
    if not np.all(rbrs[:, :1, 1] == rbrs[:, :, 1]):
        return False

    # All column's tlx equal
    if not np.all(rtls[:1, :, 0] == rtls[:, :, 0]):
        return False
    # All column's brx equal
    if not np.all(rbrs[:1, :, 0] == rbrs[:, :, 0]):
        return False

    # bounds ***********************************************
    if not allow_outer_pixels:
        if not np.all(rtls[0, 0] <= 0):
            return False
        if not np.all(rbrs[-1, -1] >= fp.rsize):
            return False
    else:
        if not np.all(rtls[0, 0] == [0, 0]):
            return False
        if not np.all(rbrs[-1, -1] == fp.rsize):
            return False

    # overlap **********************************************
    if allow_overlapping_pixels:
        # All line's consecutive tlx vs brx
        if not np.all(rbrs[:, :-1, 0] >= rtls[:, 1:, 0]):
            return False
        # All columns's consecutive tly vs bry
        if not np.all(rbrs[:-1, :, 1] >= rtls[1:, :, 1]):
            return False
    else:
        # All line's consecutive tlx vs brx
        if not np.all(rbrs[:, :-1, 0] == rtls[:, 1:, 0]):
            return False
        # All columns's consecutive tly vs bry
        if not np.all(rbrs[:-1, :, 1] == rtls[1:, :, 1]):
            return False

    return True



from typing import Set, Mapping

import collections

from buzzard._actors.message import Msg
from buzzard._footprint import Footprint
from buzzard._actors.cached.query_infos import CachedQueryInfos

class ActorCacheExtractor(object):
    """Actor that takes care of delaying reading operations according to cache state"""

    def __init__(self, raster):
        self._raster = raster
        self._alive = True

        self._path_of_cache_files_ready = {} # type: Mapping[Footprint, str]
        self._reads_waiting_for_cache_fp = (
            collections.defaultdict(lambda: collections.defaultdict(set))
        ) # type: Mapping[Footprint, Mapping[CachedQueryInfos, Set[int]]]
        self.address = '/Raster{}/CacheExtractor'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_sample_those_cache_files_to_an_array(self, qi, prod_idx):
        """Receive message: An array is needed soon as it requires one or more read operations.
        Please perform those reads when the cache files are ready.
        """
        msgs = []

        cache_fps = qi.prod[prod_idx].cache_fps
        available_cache_fps = self._path_of_cache_files_ready.keys() & cache_fps
        missing_cache_fps = cache_fps - available_cache_fps

        for cache_fp in available_cache_fps:
            msgs += [Msg(
                'Reader', 'sample_cache_file_to_unique_array',
                qi, prod_idx, cache_fp, self._path_of_cache_files_ready[cache_fp],
            )]
        for cache_fp in missing_cache_fps:
            self._reads_waiting_for_cache_fp[cache_fp][qi].add(prod_idx)

        return msgs

    def receive_cache_files_ready(self, path_of_cache_files_ready):
        """Receive message: A cache file is ready, you might already know it.

        Parameters:
        path_of_cache_files_ready: dict from Footprint to str
        """
        msgs = []

        new_cache_fps = path_of_cache_files_ready.keys() - self._path_of_cache_files_ready.keys()
        self._path_of_cache_files_ready.update(path_of_cache_files_ready)

        for cache_fp in new_cache_fps:
            # TODO Idea: Send a external message to the facade to expose the set of path to cache files with a mutex
            for qi, prod_idxs in self._reads_waiting_for_cache_fp[cache_fp].items():
                for prod_idx in prod_idxs:
                    msgs += [Msg(
                        'Reader', 'sample_cache_file_to_unique_array',
                        qi, prod_idx, cache_fp, self._path_of_cache_files_ready[cache_fp]
                    )]
            del self._reads_waiting_for_cache_fp[cache_fp]


        return msgs

    def receive_sampled_a_cache_file_to_the_array(self, qi, prod_idx, cache_fp, array):
        """Receive message: A cache file was read for that output array"""
        return [Msg(
            'Producer', 'sampled_a_cache_file_to_the_array', qi, prod_idx, cache_fp, array,
        )]

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        # Perform fine grain garbage collection
        for cache_fp in self._reads_waiting_for_cache_fp.keys() & qi.list_of_cache_fp:
            if qi in self._reads_waiting_for_cache_fp[cache_fp]:
                del self._reads_waiting_for_cache_fp[cache_fp][qi]
                if len(self._reads_waiting_for_cache_fp[cache_fp]) == 0:
                    del self._reads_waiting_for_cache_fp[cache_fp]
        return []

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False
        self._reads_waiting_for_cache_fp.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **



"""Tests for *Raster.get_data, *Raster.set_data and *Raster.fill methods (not testing resampling)"""

# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools
import os
import uuid
import tempfile

import numpy as np
import pytest

from buzzard import Footprint, Dataset

@pytest.fixture(scope='module')
def ds():
    return Dataset(allow_interpolation=0)

@pytest.fixture(scope='module', params=['GTiff', 'MEM', 'numpy'])
def driver(request):
    return request.param

@pytest.fixture(scope='module', params=[1, 3])
def channel_count(request):
    return request.param

@pytest.fixture(scope='module', params=['float32', 'uint8'])
def dtype(request):
    return request.param

@pytest.fixture(scope='module', params=[200, None])
def src_nodata(request):
    return request.param

@pytest.fixture(scope='module', params=[200, 250])
def dst_nodata(request):
    return request.param

@pytest.fixture(scope='module')
def rast(ds, driver, channel_count, dtype, src_nodata):
    """Fixture for the dataset creation"""
    fp = Footprint(
        tl=(100, 110), size=(10, 10), rsize=(10, 10)
    )
    if driver == 'numpy':
        rast = ds.awrap_numpy_raster(
            fp,
            np.empty(np.r_[fp.shape, channel_count], dtype=dtype),
            channels_schema=dict(nodata=src_nodata),
            sr=None,
            mode='w',
        )
    elif driver == 'MEM':
        rast = ds.acreate_raster(
            '', fp, dtype, channel_count, channels_schema=dict(nodata=src_nodata), driver='MEM',
        )
    else:
        path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())
        rast = ds.acreate_raster(
            path, fp, dtype, channel_count, channels_schema=dict(nodata=src_nodata), driver=driver
        )
    yield rast
    if driver in {'numpy', 'MEM'}:
        rast.close()
    else:
        rast.delete()

@pytest.fixture
def dst_arr(rast):
    """Array to write in raster and to compare against"""
    arr = np.dstack([
        np.add(*rast.fp.meshgrid_raster) + i
        for i in range(len(rast))
    ]).astype(rast.dtype)
    if rast.nodata is not None:
        arr[np.diag_indices(arr.shape[0])] = rast.nodata
    return arr

def test_fill(rast):
    for band_id in range(1, len(rast) + 1):
        rast.fill(band=band_id, value=band_id)
    for band_id in range(1, len(rast) + 1):
        assert np.all(
            rast.get_data(band=band_id) == band_id
        )

def test_set_data_whole(rast, dst_arr):
    rast.set_data(dst_arr, channels=slice(None))
    arr = rast.get_data(band=[-1])
    assert np.all(
        arr == dst_arr
    )

def test_get_data_dst_nodata(rast, dst_nodata, dst_arr):
    fp = rast.fp.dilate(1)
    inner_slice = rast.fp.slice_in(fp)
    rast.set_data(dst_arr, channels=None)

    arr = rast.get_data(band=[-1], dst_nodata=dst_nodata, fp=fp)

    arr2 = rast.get_data(band=[-1], dst_nodata=dst_nodata, fp=fp.erode(1))
    assert np.all(arr2 == arr[inner_slice])

    if rast.nodata is not None:
        assert np.all(arr[inner_slice][np.diag_indices(dst_arr.shape[0])] == dst_nodata)
        arr[inner_slice][np.diag_indices(dst_arr.shape[0])] = rast.nodata
    assert np.all(arr[inner_slice] == dst_arr)

    outer_mask = np.ones(fp.shape, bool)
    outer_mask[inner_slice] = False
    assert np.all(arr[outer_mask] == dst_nodata)

def test_get_data_rect(rast, dst_arr):
    rast.set_data(dst_arr, band=-1)

    rect_sizes = [
        1, 10, 15
    ]
    locations = [
        (-10, -10),
        (-9, -9),
        (-1, -1),
        (0, 0),
        (5, 5),
        (9, 9),
        (10, 10),
        (11, 11),
    ]
    dst_nodata = rast.nodata or 200
    for offset_px, rsize1 in itertools.product(locations, rect_sizes):
        fp = rast.fp
        fp = fp.move(fp.tl + offset_px * fp.pxvec)
        fp = fp.clip(0, 0, rsize1, rsize1)
        slice_of_file = rast.fp.slice_in(fp, clip=True)
        slice_in_file = fp.slice_in(rast.fp, clip=True)
        arr = rast.get_data(fp=fp, band=[-1], dst_nodata=dst_nodata).copy()
        if rast.fp.share_area(fp):
            assert np.all(
                dst_arr[slice_in_file] == arr[slice_of_file]
            )
            arr[slice_of_file] = dst_nodata
        assert np.all(arr == dst_nodata)

def test_set_data_rect(rast, dst_arr):
    rect_sizes = [
        1, 10, 15
    ]
    locations = [
        (-10, -10),
        (-9, -9),
        (-1, -1),
        (0, 0),
        (5, 5),
        (9, 9),
        (10, 10),
        (11, 11),
    ]
    dst_nodata = rast.nodata or 200
    for offset_px, rsize1 in itertools.product(locations, rect_sizes):
        rast.set_data(dst_arr, band=-1) # reset

        fp = rast.fp
        fp = fp.move(fp.tl + offset_px * fp.pxvec)
        fp = fp.clip(0, 0, rsize1, rsize1)
        slice_of_file = rast.fp.slice_in(fp, clip=True)
        slice_in_file = fp.slice_in(rast.fp, clip=True)

        arr1 = np.full(np.r_[fp.shape, len(rast)], 201, dtype=rast.dtype)
        rast.set_data(arr1, fp=fp, band=-1)
        arr2 = rast.get_data(band=[-1], dst_nodata=dst_nodata).copy()

        assert np.all(arr2[slice_in_file] == arr1[slice_of_file])
        arr2[slice_in_file] = dst_nodata
        assert np.all(arr2 != 201)

def test_set_data_mask(rast, dst_arr):
    fp = rast.fp.erode(rast.fp.rsemiminoraxis - 3)
    mask = np.zeros(rast.fp.shape, bool)
    mask[fp.slice_in(rast.fp)] = 1

    for mask_param in [fp.poly, mask]:
        rast.fill(band=-1, value=0)
        rast.set_data(dst_arr, band=-1, mask=mask_param)
        arr = rast.get_data(band=[-1])
        assert np.all(arr[mask] == dst_arr[mask])
        assert np.all(arr[~mask] == 0)

def test_get_data_channel_behavior(rast):
    c = len(rast)
    if c == 1:
        assert rast.get_data(band=-1).shape[2:] == ()
    else:
        assert rast.get_data(band=-1).shape[2:] == (c,)
    assert rast.get_data(band=[-1]).shape[2:] == (c,)
    assert rast.get_data(band=[1]).shape[2:] == (1,)
    assert rast.get_data(band=1).shape[2:] == ()

    if len(rast) == 3:
        for i in range(len(rast)):
            rast.fill(i * 10, i)
        assert np.all(rast.get_data(channels=[0, 1, 2]) == [[[0, 10, 20]]])
        assert np.all(rast.get_data(channels=[2, 1, 0]) == [[[20, 10, 0]]])
        assert np.all(rast.get_data(channels=[2, 1, 0, 1, 2]) == [[[20, 10, 0, 10, 20]]])



"""Conversions between gdal types python representations

http://www.gdal.org/gdal_8h.html
http://www.gdal.org/ogr__core_8h.html
"""

import datetime
import json
import collections

import shapely.wkt as sw
from osgeo import ogr, gdal
import numpy as np

# WKB (Well Known Bytes) <-> str **************************************************************** **
_WKBGEOM_KEYS = [
    key
    for key in dir(ogr)
    if key.startswith('wkb') and key not in {'wkbXDR', 'wkbNDR'}
    for val in [getattr(ogr, key)]
    if np.all(np.isreal(val))
]

_STR_OF_WKBGEOM = {
    getattr(ogr, key): key[3:]
    for key in _WKBGEOM_KEYS
}

_WKBGEOM_OF_STR = dict(
    [(v, k) for k, v in _STR_OF_WKBGEOM.items()] +
    [(v.lower(), k) for k, v in _STR_OF_WKBGEOM.items()]
)

def wkbgeom_of_str(str_):
    return _WKBGEOM_OF_STR[str_]

def str_of_wkbgeom(wkbgeom):
    return _STR_OF_WKBGEOM[wkbgeom]

# OGR Geometry <-> shapely/geojson-coordinates ************************************************** **
def ogr_of_shapely(geom):
    return ogr.CreateGeometryFromWkt(geom.wkt)

def ogr_of_coordinates(geom, type_):

    def _to_list(obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, collections.Iterable):
            return [_to_list(elt) for elt in obj]
        else:
            return obj

    d = {
        'type': type_,
        'coordinates': _to_list(geom),
    }
    d = json.dumps(d)
    d = ogr.CreateGeometryFromJson(d)
    return d

def shapely_of_ogr(geom):
    return sw.loads(geom.ExportToWkt())

def coordinates_of_ogr(geom):
    return json.loads(geom.ExportToJson())['coordinates']

# OFT (OGR Field Type) <-> type/str ************************************************************* **
# Read to parse user choices in create_vector
# Contains keys
#   - all lowercast ogr.OFT*
#   - other strings that are not interpretable by numpy
_OFT_OF_STR = {
    'binary': ogr.OFTBinary,
    'date': ogr.OFTDate,
    'datetime': ogr.OFTDateTime,
    'time': ogr.OFTTime,
    'integer': ogr.OFTInteger,
    'integer64': ogr.OFTInteger64,
    'real': ogr.OFTReal,
    'string': ogr.OFTString,
    'integer64list': ogr.OFTInteger64List,
    'int list': ogr.OFTInteger64List,
    'integerlist': ogr.OFTIntegerList,
    'reallist': ogr.OFTRealList,
    'float list': ogr.OFTRealList,
    'str list': ogr.OFTStringList,
    'stringlist': ogr.OFTStringList,
}

# Read to transform ogr.FieldDefn.type to field dict
# Contains keys
#   - All ogr.OFT*
_STR_OF_OFT = {
    ogr.OFTBinary: 'binary',
    ogr.OFTDate: 'date',
    ogr.OFTDateTime: 'datetime',
    ogr.OFTInteger64: 'integer64',
    ogr.OFTInteger64List: 'integer64list',
    ogr.OFTInteger: 'integer',
    ogr.OFTIntegerList: 'integerlist',
    ogr.OFTReal: 'real',
    ogr.OFTRealList: 'reallist',
    ogr.OFTString: 'string',
    ogr.OFTStringList: 'stringlist',
    ogr.OFTTime: 'time',
}

# Read to parse user choices in create_vector
# Contains keys
#   - All non-numpy types
#   - All numpy types
_OFT_OF_TYPE = {
    bytes: ogr.OFTBinary,
    datetime.datetime: ogr.OFTDateTime,
    np.dtype('datetime64'): ogr.OFTDateTime,
    int: ogr.OFTInteger64,
    np.dtype('int32'): ogr.OFTInteger,
    np.dtype('int64'): ogr.OFTInteger64,
    float: ogr.OFTReal,
    np.dtype('float32'): ogr.OFTReal,
    np.dtype('float64'): ogr.OFTReal,
    str: ogr.OFTString,
    np.dtype('str'): ogr.OFTString,
}

# Read to convert ogr.Field to user type
# Contains keys
#   - All ogr.OFT*
_TYPE_OF_OFT = {
    ogr.OFTBinary: np.bytes_,
    ogr.OFTDate: str,
    ogr.OFTDateTime: str,
    ogr.OFTTime: str,
    ogr.OFTInteger64: int,
    ogr.OFTInteger64List: np.int64,
    ogr.OFTInteger: int,
    ogr.OFTIntegerList: np.int32,
    ogr.OFTReal: float,
    ogr.OFTRealList: np.float64,
    ogr.OFTString: str,
    ogr.OFTStringList: np.str_,
}

def oft_of_any(obj):
    if obj in _OFT_OF_STR:
        return _OFT_OF_STR[obj]
    if obj in _OFT_OF_TYPE:
        return _OFT_OF_TYPE[obj]
    if np.dtype(obj) in _OFT_OF_TYPE:
        return _OFT_OF_TYPE[np.dtype(obj)]
    raise ValueError('Unknown type')

def str_of_oft(oft):
    return _STR_OF_OFT[oft]

def type_of_oftstr(str_):
    return _TYPE_OF_OFT[_OFT_OF_STR[str_]]

# OF (Open Flag) <-> str ************************************************************************ **
_OF_OF_STR = {
    'all': gdal.OF_ALL,
    'gnm': gdal.OF_GNM,
    'raster': gdal.OF_RASTER,
    'readonly': gdal.OF_READONLY,
    'shared': gdal.OF_SHARED,
    'update': gdal.OF_UPDATE,
    'vector': gdal.OF_VECTOR,
    'verbose_error': gdal.OF_VERBOSE_ERROR,
}

def of_of_str(str_):
    return _OF_OF_STR[str_]

_OF_OF_MODE = {
    'r': gdal.OF_READONLY,
    'w': gdal.OF_UPDATE,
}

def of_of_mode(mode):
    return _OF_OF_MODE[mode]

# GCI (GDAL Color Interpretation) <-> str ******************************************************* **
_STR_OF_GCI = {
    gdal.GCI_Undefined: 'undefined',
    gdal.GCI_GrayIndex: 'grayindex',
    gdal.GCI_PaletteIndex: 'paletteindex',
    gdal.GCI_RedBand: 'redband',
    gdal.GCI_GreenBand: 'greenband',
    gdal.GCI_BlueBand: 'blueband',
    gdal.GCI_AlphaBand: 'alphaband',
    gdal.GCI_HueBand: 'hueband',
    gdal.GCI_SaturationBand: 'saturationband',
    gdal.GCI_LightnessBand: 'lightnessband',
    gdal.GCI_CyanBand: 'cyanband',
    gdal.GCI_MagentaBand: 'magentaband',
    gdal.GCI_YellowBand: 'yellowband',
    gdal.GCI_BlackBand: 'blackband',
}

_GCI_OF_STR = {v: k for (k, v) in _STR_OF_GCI.items()}

def gci_of_str(str_):
    return _GCI_OF_STR[str_]

def str_of_gci(gci):
    return _STR_OF_GCI[gci]

# GMF (GDAL Mask Flags) <-> str ***************************************************************** **
_STR_OF_GMF = {
    gdal.GMF_ALL_VALID: 'all_valid',
    gdal.GMF_PER_DATASET: 'per_dataset',
    gdal.GMF_ALPHA: 'alpha',
    gdal.GMF_NODATA: 'nodata',
}

_GMF_OF_STR = {v: k for (k, v) in _STR_OF_GMF.items()}

def gmf_of_str(str_):
    l = str_.replace(',', ' ').split(' ')
    l = [v for v in l if len(v)]

    diff = set(l) - set(_GMF_OF_STR.keys())
    if diff:
        raise ValueError('Unknown gmf %s' % diff)
    return sum(_GMF_OF_STR[elt] for elt in l)

def str_of_gmf(gmf):
    l = []
    check = 0
    for str_, val in _GMF_OF_STR.items():
        if gmf & val:
            check |= val
            l += [str_]

    if check != gmf:
        raise ValueError('Unknown gmf bits %#x' % gmf)
    return ','.join(l)

# CPLE (Common Portability Library Errors) <-> str ********************************************** **
_STR_OF_CPLE = {
    val: key.replace('CPLE_', '')
    for key in dir(gdal)
    if key.startswith('CPLE_')
    for val in [getattr(gdal, key)]
    if isinstance(val, int)
}

_CPLE_OF_STR = {v: k for (k, v) in _STR_OF_CPLE.items()}

def cple_of_str(str_):
    return _CPLE_OF_STR[str_]

def str_of_cple(cple):
    return _STR_OF_CPLE.get(cple, 'Unknown')



from typing import Set, Dict, Tuple
import itertools
import operator
import functools
import logging
import uuid # For mypy

import sortedcontainers
import numpy as np

from buzzard._footprint import Footprint # For mypy
from buzzard._actors.message import Msg
from buzzard._actors.pool_job import PoolJobWaiting, MaxPrioJobWaiting, ProductionJobWaiting, CacheJobWaiting
from buzzard._actors.priorities import dummy_priorities, Priorities
from buzzard._actors.cached.query_infos import CachedQueryInfos

LOGGER = logging.getLogger(__name__)
OVERLOAD = 2

class ActorPoolWaitingRoom(object):
    """Actor that takes care of prioritizing jobs waiting for spots in a thread/process pool.

    It gives out tokens to allow jobs to enter the `ActorPoolWorkingRoom`. There are as many tokens
    as spots in the underlying thread/process pool.

    It accepts 3 types of `PoolJobWaiting`
    - `MaxPrioJobWaiting`
      - Rank 0 job, has priority over the other jobs.
      - Stored in a set
      - Used by `cached.FileChecker`
    - `ProductionJobWaiting`
      - Rank 1 job
      - Stored in many data structures
      - Used by `Reader`, `Resampler`, `cached.Computer`
    - `CacheJobWaiting`
      - Rank 1 job
      - Stored in many data structures
      - Used by `cached.Merger`, `cached.Writer`

    """

    def __init__(self, pool):
        """
        Parameters
        ----------
        pool: multiprocessing.pool.Pool (or the multiprocessing.pool.ThreadPool subclass)
        """
        self._alive = True

        # `global_priorities` contains all the methods necessary to establish the priority of a
        # `prod_job` or a `cache_job`. This object is updated by
        # `receive_global_priorities_update` as soon as there is an update.
        self._global_priorities = dummy_priorities # type: Priorities

        # Tokens *****************************************************
        pool_id = id(pool)
        self._pool_id = pool_id
        self._token_count = pool._processes + OVERLOAD
        short_id = short_id_of_id(pool_id)
        self._tokens = {
            # This has no particular meaning, the only hard requirement is just to have
            # different tokens in a pool.
            _PoolToken(short_id * 1000 + i)
            for i in range(self._token_count)
        }
        self._all_tokens = set(self._tokens)

        # Rank 0 jobs ************************************************
        self._jobs_maxprio = set() # type: Set[MaxPrioJobWaiting]

        # Rank 1 jobs ************************************************
        # For low complexity operations
        self._jobs_prod = set() # type: Set[ProductionJobWaiting]
        self._jobs_cache = set() # type: Set[CacheJobWaiting]

        self._dict_of_prio_per_r1job = {} # type: Dict[PoolJobWaiting, Tuple[int, ...]]
        self._sset_of_prios = sortedcontainers.SortedSet()
        self._dict_of_r1jobs_per_prio = {} # type: Dict[Tuple[int], Set[PoolJobWaiting]]

        self._prod_jobs_of_query = {} # type: Dict[CachedQueryInfos, Set[ProductionJobWaiting]]
        self._cache_jobs_of_cache_fp = {} # type: Dict[Tuple[uuid.UUID, Footprint], Set[CacheJobWaiting]]

        # Shortcuts **************************************************
        # For fast iteration / cleanup
        self._job_sets = [self._jobs_maxprio, self._jobs_prod, self._jobs_cache]
        self._data_structures = self._job_sets + [
            self._dict_of_prio_per_r1job,
            self._sset_of_prios,
            self._dict_of_r1jobs_per_prio,
            self._prod_jobs_of_query,
            self._cache_jobs_of_cache_fp,
        ]
        self.address = '/Pool{}/WaitingRoom'.format(self._pool_id)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_schedule_job(self, job):
        """Receive message: Schedule this job someday

        Parameters
        ----------
        job: _actors.pool_job.PoolJobWaiting
        """
        if len(self._tokens) != 0:
            # If job can be started straight away, do so.
            assert self._job_count == 0
            return [
                Msg(job.sender_address, 'token_to_working_room', job, self._tokens.pop())
            ]
        else:
            # Store job for later invocation
            self._store_job(job)
        return []

    def receive_unschedule_job(self, job):
        """Receive message: Forget about this waiting job

        Parameters
        ----------
        job: _actors.pool_job.PoolJobWaiting
        """
        self._unstore_job(job)
        return []

    def receive_global_priorities_update(self, global_priorities, query_updates, cache_tile_updates):
        """Receive message: Update your jobs priorities

        Parameters
        ----------
        global_priorities: Priorities
        query_updates: set of CachedQueryInfos
        cache_tile_updates: set of (raster_uid, Footprint)
        """
        # Update the version of `global_priorities`
        self._global_priorities = global_priorities

        # Update the production jobs
        for qi in query_updates & self._prod_jobs_of_query.keys():
            for job in list(self._prod_jobs_of_query[qi]):
                # Update priority
                self._unstore_job(job)
                self._store_job(job)

        # Update the cache jobs
        for raster_uid, cache_fp in cache_tile_updates & self._cache_jobs_of_cache_fp.keys():
            key = (raster_uid, cache_fp)
            for job in list(self._cache_jobs_of_cache_fp[key]):
                # Update priority
                self._unstore_job(job)
                self._store_job(job)

        return []

    def receive_salvage_token(self, token):
        """Receive message: A Job is done/cancelled, allow some other jobs

        Parameters
        ----------
        token: _PoolToken
        """
        assert token in self._all_tokens, 'Received a token that is not owned by this waiting room'
        assert token not in self._tokens, 'Received a token that is already here'
        self._tokens.add(token)

        job_count = self._job_count
        if job_count == 0:
            return []
        else:
            token_count = len(self._tokens)
            assert token_count == 1, """The way this class is designed, this point in code should only
            reached if token_count == 1"""

            job = self._unstore_most_urgent_job()

        return [Msg(
            job.sender_address, 'token_to_working_room', job, self._tokens.pop()
        )]

    def receive_die(self):
        """Receive message: The wrapped pool is no longer used"""
        assert self._alive
        self._alive = False
        if self._job_count:
            LOGGER.warning('Killing an ActorPoolWaitingRoom with {} waiting jobs'.format(
                self._job_count,
            ))

        # Clear attributes *****************************************************
        self._prios = dummy_priorities
        for ds in self._data_structures:
            ds.clear()

        return []

    # ******************************************************************************************* **
    # Misc *********************************************************************
    @property
    def _job_count(self):
        return sum(map(len, self._job_sets))

    # Job storage operations ***************************************************
    def _store_job(self, job):
        """Compute the priority of a job and register it in the right objects"""
        assert all(
            job not in set_
            for set_ in self._job_sets
        )
        if isinstance(job, MaxPrioJobWaiting):
            self._jobs_maxprio.add(job)
        else:
            if isinstance(job, ProductionJobWaiting):
                self._jobs_prod.add(job)
                if job.qi not in self._prod_jobs_of_query:
                    self._prod_jobs_of_query[job.qi] = set()
                self._prod_jobs_of_query[job.qi].add(job)
                prio = self._global_priorities.prio_of_prod_tile(job.qi, job.prod_idx)
                cx, cy = job.fp.c
                prio += (-cy, +cx, job.action_priority,)

            elif isinstance(job, CacheJobWaiting):
                self._jobs_cache.add(job)
                key = job.raster_uid, job.cache_fp
                if key not in self._cache_jobs_of_cache_fp:
                    self._cache_jobs_of_cache_fp[key] = set()
                self._cache_jobs_of_cache_fp[key].add(job)
                prio = self._global_priorities.prio_of_cache_tile(job.raster_uid, job.cache_fp)
                cx, cy = job.fp.c
                prio += (-cy, +cx, job.action_priority,)

            else: # pragma: no cover
                assert False

            self._dict_of_prio_per_r1job[job] = prio
            if prio in self._dict_of_r1jobs_per_prio:
                self._dict_of_r1jobs_per_prio[prio].add(job)
            else:
                self._dict_of_r1jobs_per_prio[prio] = {job}
                self._sset_of_prios.add(prio)

    def _unstore_job(self, job):
        """Unregister a job from the right objects"""
        if isinstance(job, MaxPrioJobWaiting):
            self._jobs_maxprio.remove(job)
        else:
            if isinstance(job, ProductionJobWaiting):
                self._jobs_prod.remove(job)
                self._prod_jobs_of_query[job.qi].remove(job)
                if len(self._prod_jobs_of_query[job.qi]) == 0:
                    del self._prod_jobs_of_query[job.qi]

            elif isinstance(job, CacheJobWaiting):
                self._jobs_cache.remove(job)
                key = job.raster_uid, job.cache_fp
                self._cache_jobs_of_cache_fp[key].remove(job)
                if len(self._cache_jobs_of_cache_fp[key]) == 0:
                    del self._cache_jobs_of_cache_fp[key]
            else: # pragma: no cover
                assert False

            prio = self._dict_of_prio_per_r1job.pop(job)
            self._dict_of_r1jobs_per_prio[prio].remove(job)
            if len(self._dict_of_r1jobs_per_prio[prio]) == 0:
                del self._dict_of_r1jobs_per_prio[prio]
                self._sset_of_prios.remove(prio)

    def _unstore_most_urgent_job(self):
        assert self._job_count > 0

        # Unstore a rank 0 job
        if len(self._jobs_maxprio) > 0:
            return self._jobs_maxprio.pop() # Pop an arbitrary one

        # Unstore a rank 1 job
        prio = self._sset_of_prios[0]
        job = next(iter(self._dict_of_r1jobs_per_prio[prio])) # Pop an arbitrary one

        self._unstore_job(job)
        return job

    # ******************************************************************************************* **

class _PoolToken(int):
    pass

def grouper(iterable, n, fillvalue=None):
    """itertools recipe: Collect data into fixed-length chunks or blocks
    grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    """
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)

def short_id_of_id(id, max_digit_count=3):
    """Shorten an integer
    Group of digits are xor'd together
    """
    id = int(id)
    it = reversed(str(id))
    it = grouper(it, max_digit_count, '0')
    it = (
        int(''.join(reversed(char_list)))
        for char_list in it
    )
    short_id = functools.reduce(operator.xor, it) % 10 ** max_digit_count
    return short_id



import numpy as np

from buzzard._a_stored_raster import AStoredRaster, ABackStoredRaster

class NumpyRaster(AStoredRaster):
    """Concrete class defining the behavior of a wrapped numpy array

    >>> help(Dataset.wrap_numpy_raster)

    Features Defined
    ----------------
    - Has an `array` property that points to the numpy array provided at construction.
    """

    def __init__(self, ds, fp, array, channels_schema, wkt, mode):
        self._arr_shape = array.shape
        self._arr_address = array.__array_interface__['data'][0]
        back = BackNumpyRaster(
            ds._back, fp, array, channels_schema, wkt, mode
        )
        super(NumpyRaster, self).__init__(ds=ds, back=back)

    @property
    def array(self):
        """Returns the Raster's full input data as a Numpy array"""
        assert (
            self._arr_address == self._back._arr.__array_interface__['data'][0]
        )
        return self._back._arr.reshape(*self._arr_shape)

class BackNumpyRaster(ABackStoredRaster):
    """Implementation of NumpyRaster"""

    def __init__(self, back_ds, fp, array, channels_schema, wkt, mode):
        array = np.atleast_3d(array)
        self._arr = array
        channel_count = array.shape[-1]

        if 'nodata' not in channels_schema:
            channels_schema['nodata'] = [None] * channel_count

        if 'interpretation' not in channels_schema:
            channels_schema['interpretation'] = ['undefined'] * channel_count

        if 'offset' not in channels_schema:
            channels_schema['offset'] = [0.] * channel_count

        if 'scale' not in channels_schema:
            channels_schema['scale'] = [1.] * channel_count

        if 'mask' not in channels_schema:
            channels_schema['mask'] = ['all_valid']

        super(BackNumpyRaster, self).__init__(
            back_ds=back_ds,
            wkt_stored=wkt,
            channels_schema=channels_schema,
            dtype=array.dtype,
            fp_stored=fp,
            mode=mode,
        )

        self._should_tranform = (
            any(v != 0 for v in channels_schema['offset']) or
            any(v != 1 for v in channels_schema['scale'])
        )

    def get_data(self, fp, channel_ids, dst_nodata, interpolation):
        samplefp = self.build_sampling_footprint(fp, interpolation)
        if samplefp is None:
            return np.full(
                np.r_[fp.shape, len(channel_ids)],
                dst_nodata,
                self.dtype
            )
        chans_indexer = self._best_indexers_of_channel_ids(channel_ids)
        key = list(samplefp.slice_in(self.fp)) + [chans_indexer]
        key = tuple(key)
        array = self._arr[key]
        if self._should_tranform:
            array = (
                array *
                np.asarray(self.channels_schema['scale'])[chans_indexer] +
                np.asarray(self.channels_schema['offset'])[chans_indexer]
            )
        array = self.remap(
            samplefp,
            fp,
            array=array,
            mask=None,
            src_nodata=self.nodata,
            dst_nodata=dst_nodata,
            mask_mode='erode',
            interpolation=interpolation,
        )
        array = array.astype(self.dtype, copy=False)
        return array

    def set_data(self, array, fp, channel_ids, interpolation, mask):
        if not fp.share_area(self.fp):
            return
        if not fp.same_grid(self.fp) and mask is None:
            mask = np.ones(fp.shape, bool)

        dstfp = self.fp.intersection(fp)

        # Remap ****************************************************************
        ret = self.remap(
            fp,
            dstfp,
            array=array,
            mask=mask,
            src_nodata=self.nodata,
            dst_nodata=self.nodata or 0,
            mask_mode='erode',
            interpolation=interpolation,
        )
        if mask is not None:
            array, mask = ret
        else:
            array = ret
        del ret
        array = array.astype(self.dtype, copy=False)
        fp = dstfp
        del dstfp

        # Write ****************************************************************
        slices = fp.slice_in(self.fp)
        for i in channel_ids:
            if mask is not None:
                self._arr[slices + (i,)][mask] = array[..., i][mask]
            else:
                self._arr[slices + (i,)] = array[..., i]

    def fill(self, value, channel_ids):
        for i in channel_ids:
            self._arr[..., i] = value

    def close(self):
        super(BackNumpyRaster, self).close()
        del self._arr

    @staticmethod
    def _best_indexers_of_channel_ids(channel_ids):
        """Create an object to pick the channels of the numpy array. Returns either a slice
        object or a list of int to perform fancy-indexing"""
        l = list(channel_ids)

        if np.all(np.diff(l) == 1):
            start, stop = l[0], l[-1] + 1
            l = slice(start, stop)
        elif np.all(np.diff(l) == -1):
            start, stop = l[0], l[-1] - 1
            if stop < 0:
                stop = None
            l = slice(start, stop, -1)
        return l



import collections

import shapely.geometry as sg

from buzzard._a_stored import AStored, ABackStored
from buzzard._a_source_vector import ASourceVector, ABackSourceVector

class AStoredVector(AStored, ASourceVector):
    """Base abstract class defining the common behavior of all vectors that are stored somewhere
    (like RAM or disk).

    Features Defined
    ----------------
    - Has an `insert_data` method that allows to write geometries to storage
    """

    def insert_data(self, geom, fields=(), index=-1):
        """.. _vector file insert_data:

        Insert a feature in vector.

        This method is not thread-safe.

        Parameters
        ----------
        geom: shapely.base.BaseGeometry or nested sequence of coordinates
            ..
        fields: sequence or dict
            Feature's fields, missing or None fields are defaulted.

            - if empty sequence: Keep all fields defaulted
            - if sequence of length len(self.fields): Fields to be set, same order as self.fields
            - if dict: Mapping of fields to be set

        index: int
            - if -1: append feature
            - else: insert feature at index (if applicable)

        Example
        -------
        >>> poly = shapely.geometry.box(10, 10, 42, 43)
        >>> fields = {'volume': 42.24}
        >>> ds.stocks.insert_data(poly, fields)

        Caveat
        ------
        When using a Vector backed by a driver (like an OGR driver), the data might be flushed to
        disk only after the garbage collection of the driver object. To be absolutely sure that the
        driver cache is flushed to disk, call `.close` or `.deactivate` on this Vector.

        """
        if geom is None: # pragma: no cover
            if not self._back.back_ds.allow_none_geometry:
                raise TypeError(
                    'Inserting None geometry not allowed '
                    '(Set `allow_none_geometry=True` in Dataset constructor to proceed)'
                )
        elif isinstance(geom, sg.base.BaseGeometry):
            geom_type = 'shapely'
        elif isinstance(geom, collections.Iterable):
            geom_type = 'coordinates'
        else:
            raise TypeError('input `geom` should be a shapely geometry or nest coordinates')
        fields = self._normalize_field_values(fields)
        self._back.insert_data(geom, geom_type, fields, index)

    def _normalize_field_values(self, fields):
        """Used on feature insertion"""
        if isinstance(fields, collections.Mapping):
            lst = [None] * len(self._back.fields)
            for k, v in fields.items():
                if v is None:
                    pass
                else:
                    i = self._back.index_of_field_name[k]
                    lst[i] = self._back.type_of_field_index[i](v)
            for defn, val in zip(self._back.fields, lst):
                if val is None and defn['nullable'] is False: # pragma: no cover
                    raise ValueError('{} not nullable'.format(defn))
            return lst
        elif isinstance(fields, collections.Iterable):
            if len(fields) == 0 and self._back.all_nullable:
                return [None] * len(self._back.fields)
            elif len(fields) != len(self._back.fields): # pragma: no cover
                raise ValueError('{} fields provided instead of {}'.format(
                    len(fields), len(self._back.fields),
                ))
            else:
                return [
                    norm(val) if val is not None else None
                    for (norm, val) in zip(self._back.type_of_field_index, fields)
                ]
        else: # pragma: no cover
            raise TypeError('Bad fields type')

class ABackStoredVector(ABackStored, ABackSourceVector):
    """Implementation of AStoredRaster's specifications"""

    def __init__(self, **kwargs):
        super(ABackStoredVector, self).__init__(**kwargs)

    def insert_data(self, geom, geom_type, fields, index): # pragma: no cover
        raise NotImplementedError('ABackStoredVector.insert_data is virtual pure')



from buzzard._a_emissary import AEmissary, ABackEmissary

class APooledEmissary(AEmissary):
    """Base abstract class defining the common behavior of all sources that can deactivate and
    reactivate their underlying driver at will.

    This is useful to balance the number of active file descriptors.
    This is useful to perform concurrent reads if the driver does no support it.

    Features Defined
    ----------------
    - An `activate` method to manually open the driver (Mostly useless feature since opening \
        is automatic if necessary)
    - A `deactivate` method to close the driver (Useful to flush data to disk)
    - An `active_count` property
    - An `active` property
    """

    def activate(self):
        """Make sure that at least one driver object is active for this Raster/Vector"""
        self._back.activate()

    def deactivate(self):
        """Collect all active driver object for this Raster/Vector. If a driver object is currently
        being used, will raise an exception."""
        self._back.deactivate()

    @property
    def active_count(self):
        """Count how many driver objects are currently active for this Raster/Vector"""
        return self._back.active_count

    @property
    def active(self):
        """Is there any driver object currently active for this Raster/Vector"""
        return self._back.active

class ABackPooledEmissary(ABackEmissary):
    """Implementation of APooledEmissary"""

    def __init__(self, uid, **kwargs):
        self.uid = uid
        super(ABackPooledEmissary, self).__init__(**kwargs)

    def activate(self):
        self.back_ds.activate(self.uid, self.allocator)

    def deactivate(self):
        self.back_ds.deactivate(self.uid)

    @property
    def active_count(self):
        return self.back_ds.active_count(self.uid)

    @property
    def active(self):
        return self.back_ds.active_count(self.uid) > 0

    def close(self):
        """Virtual method:
        - May be overriden
        - Should always be called
        """
        self.back_ds.deactivate(self.uid)
        super(ABackPooledEmissary, self).close()

    def delete(self):
        """Virtual method:
        - May be overriden
        - Should always be called

        Quick hotfix to allow `.delete` to work on Windows in tempdir.
        Ideally a `close` should be performed before a delete.
        """
        self.back_ds.deactivate(self.uid)
        super(ABackPooledEmissary, self).delete()



import functools
import collections
import multiprocessing as mp
import multiprocessing.pool

import numpy as np

from buzzard._actors.message import Msg
from buzzard._actors.pool_job import CacheJobWaiting, PoolJobWorking

class ActorMerger(object):
    """Actor that takes care of merging several arrays into one fp"""

    def __init__(self, raster):
        self._raster = raster
        self._alive = True
        merge_pool = raster.merge_pool
        if merge_pool is not None:
            self._waiting_room_address = '/Pool{}/WaitingRoom'.format(id(merge_pool))
            self._working_room_address = '/Pool{}/WorkingRoom'.format(id(merge_pool))
            if isinstance(merge_pool, mp.pool.ThreadPool):
                self._same_address_space = True
            elif isinstance(merge_pool, mp.pool.Pool):
                self._same_address_space = False
            else: # pragma: no cover
                assert False, 'Type should be checked in facade'
        self._waiting_jobs = set()
        self._working_jobs = set()

        self.dst_array = None
        self.address = '/Raster{}/Merger'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_merge_those_arrays(self, cache_fp, array_per_fp):
        msgs = []
        assert len(array_per_fp) > 0

        if len(array_per_fp) == 1:
            (fp, arr), = array_per_fp.items()
            assert fp.almost_equals(cache_fp)
            msgs += [
                Msg('Writer', 'write_this_array', cache_fp, arr)
            ]
        elif self._raster.merge_pool is None:
            work = self._create_work_job(cache_fp, array_per_fp)
            res = work.func()
            res = self._normalize_user_result(cache_fp, res)
            msgs += self._commit_work_result(work, res)
        else:
            wait = Wait(self, cache_fp, array_per_fp)
            self._waiting_jobs.add(wait)
            msgs += [Msg(self._waiting_room_address, 'schedule_job', wait)]

        return msgs

    def receive_token_to_working_room(self, job, token):
        self._waiting_jobs.remove(job)
        work = self._create_work_job(job.cache_fp, job.array_per_fp)
        self._working_jobs.add(work)
        return [
            Msg(self._working_room_address, 'launch_job_with_token', work, token)
        ]

    def receive_job_done(self, job, result):
        self._working_jobs.remove(job)
        return self._commit_work_result(job, result)

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        msgs = []
        for job in self._waiting_jobs:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
        for job in self._working_jobs:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
        self._waiting_jobs.clear()
        self._working_jobs.clear()
        self._raster = None

        return msgs

    # ******************************************************************************************* **
    def _create_work_job(self, cache_fp, array_per_fp):
        return Work(
            self, cache_fp, array_per_fp
        )

    def _commit_work_result(self, job, arr):
        return [
            Msg('Writer', 'write_this_array', job.cache_fp, arr)
        ]

    def _normalize_user_result(self, cache_fp, res):
        try:
            res = np.atleast_3d(res)
        except: # pragma: no cover
            raise ValueError("Result of recipe's `merge_arrays` has type {}, it can't be converted to ndarray".format(
                type(res)
            ))
        y, x, c = res.shape
        if (y, x) != tuple(cache_fp.shape): # pragma: no cover
            raise ValueError("Result of recipe's `merge_arrays` has shape `{}`, should start with {}".format(
                res.shape,
                cache_fp.shape,
            ))
        if c != len(self._raster): # pragma: no cover
            raise ValueError("Result of recipe's `merge_arrays` has shape `{}`, should have {} bands".format(
                res.shape,
                len(self._raster),
            ))
        res = res.astype(self._raster.dtype, copy=False)
        return res

    # ******************************************************************************************* **

class Wait(CacheJobWaiting):
    def __init__(self, actor, cache_fp, array_per_fp):
        self.cache_fp = cache_fp
        self.array_per_fp = array_per_fp
        super().__init__(actor.address, actor._raster.uid, self.cache_fp, 3, self.cache_fp)

class Work(PoolJobWorking):
    def __init__(self, actor, cache_fp, array_per_fp):
        self.cache_fp = cache_fp

        if actor._raster.merge_pool is None or actor._same_address_space:
            func = functools.partial(
                actor._raster.merge_arrays,
                cache_fp,
                array_per_fp,
                actor._raster.facade_proxy,
            )
        else:
            func = functools.partial(
                actor._raster.merge_arrays,
                cache_fp,
                array_per_fp,
                None
            )
        actor._raster.debug_mngr.event('object_allocated', func)

        super().__init__(actor.address, func)



# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import logging
import collections
import itertools
import tempfile
import os
import uuid

import numpy as np
import pytest
import shapely.geometry as sg
from osgeo import gdal

import buzzard as buzz
from buzzard.test import make_tile_set
from buzzard.test.tools import eqall
from .tools import SRS

LOGGER = logging.getLogger('buzzard')

@pytest.fixture(scope='module')
def fps():
    """
    len(fps) = 225
    See make_tile_set
    A B C D E
    F G H I J
    K L M N O
    P Q R S T
    U V W X Y
    """
    return make_tile_set.make_tile_set(5, [0.1, -0.1])

def pytest_generate_tests(metafunc):
    if 'test_fields' in metafunc.fixturenames:
        metafunc.parametrize(
            argnames='driver,suffix,test_fields,test_coords_insertion',
            argvalues=[
                ('ESRI Shapefile', '.shp', True, True), # 145% time
                ('GeoJson', '.json', True, True), # 100% time
                ('Memory', '', True, True),
                # ('BNA', '.bna', False, True), # 350% time
                # ('DGN', '.dgn', False, False), # 190% time
                ('DXF', '.dxf', False, False), # 490% time
            ],
        )
    else:
        metafunc.parametrize(
            argnames='driver,suffix',
            argvalues=[
                ('ESRI Shapefile', '.shp'),
            ],
        )

@pytest.fixture()
def path(suffix, driver):
    """Create a temporary path, and take care of cleanup afterward"""
    path = '{}/{}{}'.format(tempfile.gettempdir(), uuid.uuid4(), suffix)
    yield path
    if os.path.isfile(path):
        try:
            dr = gdal.GetDriverByName(driver)
            dr.Delete(path)
        except:
            os.remove(path)


def test_run(path, driver, fps, test_fields, test_coords_insertion):
    # Step 1 - Build file according to fixture parameters **********************
    ds = buzz.Dataset()

    if test_fields:
        fields = FIELDS
    else:
        fields = []

    geom_type = 'polygon'
    v = ds.create_avector(path, geom_type, fields, driver=driver, sr=SRS[0]['wkt'])

    def _build_test_data():
        """Create test data of random format"""
        rng = np.random.RandomState(42)
        for fpname, fp in fps.items():
            geom = _geom_of_fp(rng, fp, test_coords_insertion)

            # Keep `_fields_of_fp` invocation before `if` for rng side effect
            fields = _fields_of_fp(rng, fp, fpname)

            if test_fields:
                yield geom, fields
            else:
                yield geom,

    data = list(_build_test_data())
    for dat in data:
        v.insert_data(*dat)
    if driver in {'GeoJson', 'DXF'}:
        # Flushing to disk for geojson
        v.deactivate()
        v.activate()

    # Step 2 - Test geometries read routines ***********************************
    _test_geom_read(v, fps, data, test_fields)

    # Step 3 - Test field read routines ****************************************
    if test_fields:
        _test_fields_read(v, data)

    v.close()

# Depth 1 - Write subroutines ******************************************************************* **
FIELDS = [
    {'name': 'rarea', 'type': int},
    {'name': 'fpname', 'type': str},
    {'name': 'sqrtarea', 'type': float},
]

def _geom_of_fp(rng, fp, test_coords_insertion):
    poly = fp.poly
    i = rng.randint(2)
    if not test_coords_insertion:
        return poly
    if i:
        return poly
    else:
        coords = sg.mapping(poly)['coordinates']
        if rng.randint(2):
            coords = np.asarray(coords)
        return coords

def _fields_of_fp(rng, fp, fpname):
    full = [
        ('rarea', fp.rarea),
        ('fpname', fpname),
        ('sqrtarea', fp.area ** .5),
    ]

    if rng.randint(2):
        # Fields in a `dict`
        i = rng.randint(4)
        if i == 0:
            # Dict - All present
            return dict(full)
        elif i == 1:
            # Dict - All absent
            return {}
        elif i == 2:
            # Dict - All None
            return {field_def['name']: None for field_def in FIELDS}
        else:
            # Dict - Per field
            assert i == 3
            d = {}
            for k, v in full:
                i = rng.randint(3)
                if i == 0:
                    d[k] = v
                elif i == 1:
                    d[k] = None
                else:
                    assert i == 2
            return d
    else:
        # Fields in a `list`
        i = rng.randint(4)
        if i == 0:
            # List - All present
            return [v for (_, v) in full]
        elif i == 1:
            # List - All absent
            return []
        elif i == 2:
            # List - All None
            return [None] * len(full)
        else:
            # List - Per field
            assert i == 3
            l = []
            for _, v in full:
                if rng.randint(2):
                    l.append(v)
                else:
                    l.append(None)
            return l

# Depth 1 - Read subroutines ******************************************************************** **
def _test_fields_read(v, data):
    """Test fields reading with iter_data

    Not testing geojson, no testing get_* functions
    Not actually testing fields value,
    - testing that all test yield the same thing
    - testing that all lengths are ok
    """

    query_waysssss = [
        [[], '', None],
        [[0], 'rarea', ['rarea']],
        [[1], 'fpname', ['fpname']],
        [[2], 'sqrtarea', ['sqrtarea']],
        [[0, 1], 'rarea,fpname', ['rarea', 'fpname'], [0, 'fpname'], ['rarea', 1]],
        [[0, 2], 'rarea,sqrtarea', ['rarea', 'sqrtarea'], [0, 'sqrtarea'], ['rarea', 2]],
        [[1, 0], 'fpname,rarea', ['fpname', 'rarea'], [1, 'rarea'], ['fpname', 0]],
        [[1, 2], 'fpname,sqrtarea', ['fpname', 'sqrtarea'], [1, 'sqrtarea'], ['fpname', 2]],
        [[2, 1], 'sqrtarea,fpname', ['sqrtarea', 'fpname'], [2, 'fpname'], ['sqrtarea', 1]],
        [[2, 0], 'sqrtarea,rarea', ['sqrtarea', 'rarea'], [2, 'rarea'], ['sqrtarea', 0]],
        [[0, 1, 2], -1, ['rarea', 'fpname', 'sqrtarea'], [0, 'fpname', 'sqrtarea'], ['rarea', 1, 2]],
    ]

    # ds = buzz.Dataset()
    # v = ds.open_vector('v', path, driver=driver)

    for query_ways in query_waysssss:
        # All queries in `query_ways` request the same thing in a different way
        indices = query_ways[0]

        queries_results = []
        for query in query_ways:
            features = list(v.iter_data(query))
            assert len(features) == len(data)

            query_result = []
            for feature in features:
                # For each feature in the polygon
                if len(indices) == 0:
                    feature = [feature]

                fields = feature[1:]
                assert len(fields) == len(indices)

                query_result.append(fields) # Build list of fields
            queries_results.append(query_result) # Build list of list of fields
            del query_result

        assert len(queries_results) == len(query_ways)
        _assert_all_list_of_fields_same(queries_results)

def _test_geom_read(v, fps, data, test_fields):
    """Test many combinations of parameters for iter/get_data/geojson. Only check geometry"""
    # ds = buzz.Dataset()
    # v = ds.open_vector('v', path, driver=driver)

    queries = _build_geom_read_queries(v, fps, test_fields)
    for gen, slicing, _, mask, clip in queries:
        # Normalize input mask
        if mask is None:
            pass
        elif isinstance(mask, buzz.Footprint):
            mask = mask.poly
        elif isinstance(mask, sg.Polygon):
            pass
        elif isinstance(mask, (tuple, np.ndarray)):
            mask = sg.box(*np.asarray(mask)[[0, 2, 1, 3]])
        else:
            print('fail with type', type(mask))
            assert False

        # Normalize input geometry
        geoms_ref = [dat[0] for dat in data]
        geoms_ref = [_any_geom_to_shapely(geom) for geom in geoms_ref]
        if mask is not None:
            geoms_ref = [
                geom for geom in geoms_ref
                if not geom.disjoint(mask)
            ]
        if clip:
            geoms_ref = [
                geom & mask
                for geom in geoms_ref
            ]
        geoms_ref = geoms_ref[slicing]

        # Compare
        geoms = list(gen)
        geoms = [_any_geom_to_shapely(geom) for geom in geoms]
        geoms_ref = [_any_geom_to_shapely(geom) for geom in geoms_ref]
        assert len(geoms) == len(geoms_ref)
        for geom, geom_ref in zip(geoms, geoms_ref):
            assert (geom ^ geom_ref).is_empty

# Depth 2 - Read subroutines ******************************************************************** **
def _build_geom_read_queries(v, fps, test_fields):
    slicings = [
        slice(None),
        slice(1, None, 1),
        slice(None, 5, 1),
        slice(1, 5, 1),
        slice(1, None, 2),
        slice(None, 5, 2),
        slice(1, 5, 2),
    ]
    geom_types = ['shapely', 'coordinates']
    masks = [
        None,
        fps.GS,
        fps.GS.poly,
        fps.GS.extent,
    ]
    clips = [True, False]

    def _iter_by_get_data(geom_type, slicing, mask, clip):
        try:
            for i in range(*slicing.indices(len(v))):
                yield v.get_data(i, None, geom_type, mask, clip)
        except IndexError:
            pass

    def _iter_by_get_geojson(slicing, mask, clip):
        try:
            for i in range(*slicing.indices(len(v))):
                yield v.get_geojson(i, mask, clip)
        except IndexError:
            pass

    prod = itertools.product
    queries = [
        ((v.iter_data(None, geom_type, mask, clip, slicing), slicing, geom_type, mask, clip)
         for geom_type, slicing, mask, clip in prod(geom_types, slicings, masks, clips)
         if not (clip is True and mask is None)
        ),
        ((_iter_by_get_data(geom_type, slicing, mask, clip), slicing, geom_type, mask, clip)
         for geom_type, slicing, mask, clip in prod(geom_types, slicings, masks, clips)
         if not (clip is True and mask is None)
        ),
    ]
    if test_fields:
        queries += [
            ((v.iter_geojson(mask, clip, slicing), slicing, 'geojson', mask, clip)
             for slicing, mask, clip in prod(slicings, masks, clips)
             if not (clip is True and mask is None)
            ),

            ((_iter_by_get_geojson(slicing, mask, clip), slicing, 'geojson', mask, clip)
             for slicing, mask, clip in prod(slicings, masks, clips)
             if not (clip is True and mask is None)
            ),
        ]

    queries = itertools.chain(*queries)
    return queries

def _any_geom_to_shapely(geom):
    """Any geom to shapely object. Points should have homogeneous dimensions size."""
    if isinstance(geom, (sg.LineString, sg.Point, sg.Polygon, sg.MultiPolygon)):
        return geom
    if isinstance(geom, dict):
        return sg.shape(geom['geometry'])
    if isinstance(geom, collections.Container):
        geom = np.asarray(geom)
        if geom.ndim == 1:
            return sg.Point(geom.tolist())
        elif geom.ndim == 2:
            return sg.LineString(geom.tolist())
        elif geom.ndim == 3:
            return sg.Polygon(*geom.tolist())
        elif geom.ndim == 4:
            return sg.MultiPolygon([
                sg.Polygon(*poly)
                for poly in geom.tolist()
            ])
    assert False

def _assert_all_list_of_fields_same(llf):
    """Check that all queries in query_ways yielded the same result"""
    fields_count = len(llf[0])

    assert eqall([len(lf) for lf in llf])

    for i in range(fields_count):
        val = llf[0][i]
        assert all(val == lf[i] for lf in llf)

def test_value_error(path):
    ds = buzz.Dataset()
    v = ds.acreate_vector(path, type='polygon')

    with pytest.raises(ValueError, match='geom_type'):
        list(v.iter_data(geom_type=''))
    with pytest.raises(TypeError, match='slicing'):
        list(v.iter_data(slicing=0))
    with pytest.raises(ValueError, match='clip'):
        list(v.iter_data(clip=True))
    with pytest.raises(TypeError, match='slicing'):
        list(v.iter_geojson(slicing=0))
    with pytest.raises(ValueError, match='clip'):
        list(v.iter_geojson(clip=True))
    with pytest.raises(TypeError, match='a'):
        v.insert_data(42)

def test_iter_data_fields_behavior(path):
    ds = buzz.Dataset()
    with ds.acreate_vector(path, fields=[], type='point').remove as v:
        pt0, pt1 = sg.Point(1, 1), sg.Point(2, 2)
        v.insert_data(pt0)
        v.insert_data(pt1)
        assert list(v.iter_data(None)) == [
            pt0, pt1
        ]
        assert list(v.iter_data(-1)) == [
            pt0, pt1
        ]
        assert list(v.iter_data('')) == [
            (pt0,), (pt1,)
        ]
        assert list(v.iter_data([])) == [
            (pt0,), (pt1,)
        ]
        assert list(v.iter_data([-1])) == [
            (pt0,), (pt1,)
        ]
    with ds.acreate_vector(path, fields=[dict(name='toto', type=int)], type='point').remove as v:
        pt0, pt1 = sg.Point(1, 1), sg.Point(2, 2)
        v.insert_data(pt0, [42])
        v.insert_data(pt1, [43])
        assert list(v.iter_data(None)) == [
            pt0, pt1
        ]
        assert list(v.iter_data(-1)) == [
            (pt0, 42), (pt1, 43)
        ]
        assert list(v.iter_data('')) == [
            (pt0,), (pt1,)
        ]
        assert list(v.iter_data([])) == [
            (pt0,), (pt1,)
        ]
        assert list(v.iter_data([-1])) == [
            (pt0, 42), (pt1, 43)
        ]



from buzzard._a_stored import AStored, ABackStored

from buzzard import _tools

class AEmissary(AStored):
    """Base abstract class defining the common behavior of all sources that are backed by a driver.

    Features Defined
    ----------------
    - Has a `driver` (like "GTiff" for GDAL's geotiff driver)
    - Has `open_options`
    - Has a `path` (if the driver supports it)
    - Can be deleted (if the driver supports it)
    """

    @property
    def driver(self):
        """Get the driver name, such as 'GTiff' or 'GeoJSON'"""
        return self._back.driver

    @property
    def open_options(self):
        """Get the list of options used for opening"""
        return self._back.open_options

    @property
    def path(self):
        """Get the file system path of this source, may be the empty string if not applicable"""
        return self._back.path

    @property
    def delete(self):
        """Delete a source with a call or a context management. May raise an exception if not
        applicable or if `mode` = 'r'
        The `delete` attribute returns an object that can be both called and used in a with statement

        Example
        -------
        >>> ds.dem.delete()
        >>> with ds.dem.delete:
                # code...
        >>> with ds.acreate_raster('/tmp/tmp.tif', fp, float, 1).delete as tmp:
                # code...
        >>> with ds.acreate_vector('/tmp/tmp.shp', 'polygon').delete as tmp:
                # code...
        """
        if self.mode != 'w':
            raise RuntimeError('Cannot remove a read-only file')

        def _delete():
            self._back.delete()
            self.close()

        return _DeleteRoutine(self, _delete)

    remove = delete

class ABackEmissary(ABackStored):
    """Implementation of AEmissary's specifications"""

    def __init__(self, driver, open_options, path, **kwargs):
        self.driver = driver
        self.open_options = open_options
        self.path = path
        super(ABackEmissary, self).__init__(**kwargs)


    def delete(self):
        """Virtual method:
        - May be overriden
        - Should always be called
        """
        pass

_DeleteRoutine = type('_DeleteRoutine', (_tools.CallOrContext,), {
    '__doc__': AEmissary.delete.__doc__,
})



from buzzard._actors.message import Msg

class ActorProductionGate(object):
    """Actor that takes care of delaying the production of arrays until needed"""

    def __init__(self, raster):
        self._raster = raster
        self._queries = {}
        self._alive = True
        self.address = '/Raster{}/ProductionGate'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_make_those_arrays(self, qi):
        """Receive message: New query.

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        msgs = []
        assert qi not in self._queries

        q = _Query()
        self._queries[qi] = q
        msgs += self._allow(qi, q, 0)
        return msgs

    def receive_output_queue_update(self, qi, produced_count, queue_size):
        """Receive message: The output queue of a query changed in size.

        If necessary, allow some production array contruction to start

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        produced_count: int
            How many arrays were pushed in the output queue
        queue_size: int
            How many arrays are currently available in the output queue
        """
        msgs = []

        q = self._queries[qi]
        if produced_count == qi.produce_count:
            assert q.allowed_count == produced_count
            del self._queries[qi]
        else:
            pulled_count = produced_count - queue_size
            msgs += self._allow(qi, q, pulled_count)

        return msgs

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        del self._queries[qi]
        return []

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        self._queries.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **
    @staticmethod
    def _allow(qi, q, pulled_count):
        msgs = []

        while True:
            if q.allowed_count == qi.produce_count:
                # All productions started
                break
            if q.allowed_count == pulled_count + qi.max_queue_size:
                # Enough production started yet
                break
            msgs += [Msg(
                'Producer', 'make_this_array', qi, q.allowed_count
            )]
            q.allowed_count += 1

        return msgs


    # ******************************************************************************************* **

class _Query(object):

    def __init__(self):
        self.allowed_count = 0



import sys

import numpy as np

from buzzard._a_source import ASource, ABackSource
from buzzard._a_source_raster_remap import ABackSourceRasterRemapMixin
from buzzard._footprint import Footprint
from buzzard import _tools

class ASourceRaster(ASource):
    """Base abstract class defining the common behavior of all rasters.

    Features Defined
    ----------------
    - Has a `stored` Footprint that defines the location of the raster
    - Has a Footprint that is influenced by the Dataset's opening mode
    - Has a length that defines how many channels are available
    - Has a `channels_schema` that defines per channel attributes (e.g. nodata)
    - Has a `dtype` (like np.float32)
    - Has a `get_data` method that allows to read pixels in their current state to numpy arrays
    """

    @property
    def fp_stored(self):
        return self._back.fp_stored

    @property
    def fp(self):
        return self._back.fp

    @property
    def channels_schema(self):
        return dict(self._back.channels_schema)

    @property
    def dtype(self):
        return self._back.dtype

    @property
    def nodata(self):
        """Accessor for first channel's nodata value"""
        return self._back.nodata

    def get_nodata(self, channel=0):
        """Accessor for nodata value"""
        return self._back.get_nodata(channel)

    def __len__(self):
        """Return the number of channels"""
        return len(self._back)

    def get_data(self, fp=None, channels=None, dst_nodata=None, interpolation='cv_area', **kwargs):
        """.. _raster file get_data:

        Read a rectangle of data on several channels from the source raster.

        If `fp` is not fully within the source raster, the external pixels are set to nodata. If
        nodata is missing, 0 is used.
        If `fp` is not on the same grid as the source raster, remapping is performed using
        `interpolation` algorithm. (It fails if the `allow_interpolation` parameter is set to
        False in `Dataset` (default)). When remapping, the nodata values are not interpolated,
        they are correctly spread to the output.

        If `dst_nodata` is provided, nodata pixels are set to `dst_nodata`.

        .. warning::
            The alpha channels are currently resampled like any other channels, this behavior may
            change in the future. To normalize an `rgba` array after a resampling operation, use this
            piece of code:

            >>> arr = np.where(arr[..., -1] == 255, arr, 0)

        .. warning::
            Bands in GDAL are indexed from 1. Channels in buzzard are indexed from 0.

        Parameters
        ----------
        fp: Footprint of shape (Y, X) or None
            If None: return the full source raster

            If Footprint: return this window from the raster
        channels: None or int or slice or sequence of int (see `Channels Parameter` below)
            The channels to be read
        dst_nodata: nbr or None
            nodata value in output array
            If None and raster.nodata is not None: raster.nodata is used
            If None and raster.nodata is None: 0 is used
        interpolation: one of {'cv_area', 'cv_nearest', 'cv_linear', 'cv_cubic', 'cv_lanczos4'} or None
            OpenCV method used if intepolation is necessary

        Returns
        -------
        array: numpy.ndarray of shape (Y, X) or (Y, X, C)
            - If the `channels` parameter is `-1`, the returned array is of shape (Y, X) when `C=1`, \
               (Y, X, C) otherwise.
            - If the `channels` parameter is an integer `>=0`, the returned array is of shape (Y, X).
            - If the `channels` parameter is a sequence or a slice, the returned array is always of\
               shape (Y, X, C), no matter the size of `C`.

            (see :ref:`Channels Parameter` below)

        .. _Channels Parameter:
        Channels Parameter
        ------------------

        +------------+-----------------------------------------------------+----------------+---------------------+
        | type       | value                                               | meaning        | output shape        |
        +============+=====================================================+================+=====================+
        | NoneType   | None (default)                                      | All channels   | (Y, X) or (Y, X, C) |
        +------------+-----------------------------------------------------+----------------+---------------------+
        | slice      | slice(None), slice(1), slice(0, 2), slice(2, 0, -1) | Those channels | (Y, X, C)           |
        +------------+-----------------------------------------------------+----------------+---------------------+
        | int        | 0, 1, 2, -1, -2, -3                                 | Channel `idx`  | (Y, X)              |
        +------------+-----------------------------------------------------+----------------+---------------------+
        | (int, ...) | [0], [1], [2], [-1], [-2], [-3], [0, 1], [-1, 2, 1] | Those channels | (Y, X, C)           |
        +------------+-----------------------------------------------------+----------------+---------------------+

        """
        dst_nodata, kwargs = _tools.deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='dst_nodata', old_names={'nodata': '0.5.0'}, context='ASourceRaster.get_data',
            new_name_value=dst_nodata,
            new_name_is_provided=dst_nodata != None,
            user_kwargs=kwargs,
        )

        def _band_to_channels(val):
            val = np.asarray(val)
            if np.array_equal(val, -1):
                return None
            if val.ndim == 0:
                return val - 1
            if val.ndim != 1:
                raise ValueError('Error in deprecated `band` parameter')
            val = [
                v
                for v in val
                for v in (range(len(self)) if v == -1 else [v - 1])
            ]
            return val
        channels, kwargs = _tools.deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='channels', old_names={'band': '0.6.0'}, context='ASourceRaster.get_data',
            new_name_value=channels,
            new_name_is_provided=channels is not None,
            user_kwargs=kwargs,
            transform_old=_band_to_channels,
        )
        if kwargs: # pragma: no cover
            raise TypeError("get_data() got an unexpected keyword argument '{}'".format(
                list(kwargs.keys())[0]
            ))

        # Normalize and check fp parameter
        if fp is None:
            fp = self.fp
        elif not isinstance(fp, Footprint): # pragma: no cover
            raise ValueError('`fp` parameter should be a Footprint (not {})'.format(fp))

        # Normalize and check channels parameter
        channel_ids, is_flat = _tools.normalize_channels_parameter(
            channels, len(self)
        )
        if is_flat:
            outshape = tuple(fp.shape)
        else:
            outshape = tuple(fp.shape) + (len(channel_ids),)
        del channels

        # Normalize and check dst_nodata parameter
        if dst_nodata is not None:
            dst_nodata = self.dtype.type(dst_nodata)
        elif self.nodata is not None:
            dst_nodata = self.nodata
        else:
            dst_nodata = self.dtype.type(0)

        # Check interpolation parameter
        if not (interpolation is None or
                interpolation in self._back.REMAP_INTERPOLATIONS): # pragma: no cover
            raise ValueError('`interpolation` should be None or one of {}'.format(
                set(self._back.REMAP_INTERPOLATIONS.keys())
            ))

        return self._back.get_data(
            fp=fp,
            channel_ids=channel_ids,
            dst_nodata=dst_nodata,
            interpolation=interpolation,
        ).reshape(outshape)

    # Deprecation
    fp_origin = _tools.deprecation_pool.wrap_property(
        'fp_stored',
        '0.4.4'
    )

    band_schema = _tools.deprecation_pool.wrap_property(
        'channels_schema',
        '0.6.0'
    )

class ABackSourceRaster(ABackSource, ABackSourceRasterRemapMixin):
    """Implementation of ASourceRaster's specifications"""

    def __init__(self, channels_schema, dtype, fp_stored, **kwargs):
        super(ABackSourceRaster, self).__init__(rect=fp_stored, **kwargs)

        if self.to_work is not None:
            fp = fp_stored.move(*self.to_work([
                fp_stored.tl, fp_stored.tr, fp_stored.br
            ]), round_coordinates=True)
        else:
            fp = fp_stored

        self.channels_schema = channels_schema
        self.dtype = dtype
        self.fp_stored = fp_stored

        self.fp = fp

    @property
    def nodata(self):
        return self.get_nodata(0)

    def get_nodata(self, channel=0):
        return self.channels_schema['nodata'][channel]

    def __len__(self):
        return len(self.channels_schema['nodata'])

    def get_data(self, fp, channels, dst_nodata, interpolation): # pragma: no cover
        raise NotImplementedError('ABackSourceRaster.get_data is virtual pure')

if sys.version_info < (3, 6):
    # https://www.python.org/dev/peps/pep-0487/
    for k, v in ASourceRaster.__dict__.items():
        if hasattr(v, '__set_name__'):
            v.__set_name__(ASourceRaster, k)



""">>> help(buzz.env)
>>> help(buzz.Env)
"""

import threading
from collections import namedtuple
import functools

import cv2
from osgeo import gdal, ogr, osr

from buzzard._tools import conv, Singleton, deprecation_pool

try:
    from collections import ChainMap
except:
    # https://pypi.python.org/pypi/chainmap
    from chainmap import ChainMap

# Sanitization ********************************************************************************** **
_INDEX_DTYPES = list(conv.DTYPE_OF_NAME.keys())
def _sanitize_index_dtype(val):
    val = conv.dtype_of_any_downcast(val)
    if val not in _INDEX_DTYPES:
        raise ValueError('%s cannot be used as an index dtype' % val)
    return val

def _sanitize_significant(val):
    val = float(val)
    if val <= 0:
        raise ValueError('Significant should be greater than 0')
    return val

# Options declaration *************************************************************************** **
_EnvOption = namedtuple('_Option', 'sanitize, set_up, bottom_value')
_OPTIONS = {
    'significant': _EnvOption(_sanitize_significant, None, 9.0),
    'default_index_dtype': _EnvOption(_sanitize_index_dtype, None, 'int32'),
    'allow_complex_footprint': _EnvOption(bool, None, False),
}

# Storage *************************************************************************************** **
class _GlobalMapStack:
    """ChainMap updated to behave like a singleton stack"""

    _main_storage = None

    def __init__(self, bottom=None):
        if bottom is not None:
            # Create bottom
            self._mapping = ChainMap(bottom)
            assert self.__class__._main_storage is None
            self.__class__._main_storage = self
        else:
            # Retrieve bottom from main thread and perform a deep copy
            assert self.__class__._main_storage is not None
            self._mapping = ChainMap(*[
                dict(mapping)
                for mapping in self._main_storage._mapping.maps
            ])

    def push(self, mapping):
        self._mapping = self._mapping.new_child(mapping)

    def remove_top(self):
        assert len(self._mapping.parents.maps) > 0
        self._mapping = self._mapping.parents

    def __getitem__(self, k):
        return self._mapping[k]

class _Storage(threading.local):
    """Thread local storage for the GlobalMapStack instance"""
    def __init__(self):
        if threading.current_thread().__class__.__name__ == '_MainThread':
            self._mapstack = _GlobalMapStack({
                k: v.sanitize(v.bottom_value)
                for k, v in _OPTIONS.items()
            })
        else:
            self._mapstack = _GlobalMapStack()
        threading.local.__init__(self)

_LOCAL = _Storage()

# Env update ************************************************************************************ **
class Env(object):
    """Context manager to update buzzard's states. Can also be used as a decorator.

    Parameters
    ----------
    significant: int
        Number of significant digits for floating point comparisons
        Initialized to `9.0`
        see: https://github.com/earthcube-lab/buzzard/wiki/Precision-system
        see: https://github.com/earthcube-lab/buzzard/wiki/Floating-Point-Considerations
    default_index_dtype: convertible to np.dtype
        Default numpy return dtype for array indices.
        Initialized to `np.int32` (signed to allow negative indices by default)
    allow_complex_footprint: bool
        Whether to allow non north-up / west-left Footprints
        Initialized to `False`

    Examples
    --------
    >>> import buzzard as buzz
    >>> with buzz.Env(default_index_dtype='uint64'):
    ...     ds = buzz.Dataset()
    ...     dsm = ds.aopen_raster('dsm', 'path/to/dsm.tif')
    ...     x, y = dsm.meshgrid_raster
    ...     print(x.dtype)
    numpy.uint64

    >>> @buzz.Env(allow_complex_footprint=True)
    ... def main():
    ...     fp = buzz.Footprint(rsize=(10, 10), gt=(100, 1, 0, 100, 0, 1))

    """

    def __init__(self, **kwargs):
        kwargs = deprecation_pool.handle_param_removal_with_kwargs(
            {'warnings': '0.6.0'}, 'Env', kwargs,
        )
        self._mapping = {}
        for k, v in kwargs.items():
            if k not in _OPTIONS: # pragma: no cover
                raise ValueError('Unknown env key')
            v = _OPTIONS[k].sanitize(v)
            self._mapping[k] = v

    def __enter__(self):
        for k, newv in self._mapping.items():
            if _OPTIONS[k].set_up is not None:
                oldv = _LOCAL._mapstack[k]
                _OPTIONS[k].set_up(newv, oldv)
        _LOCAL._mapstack.push(self._mapping)

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        _LOCAL._mapstack.remove_top()
        for k, oldv in self._mapping.items():
            if _OPTIONS[k].set_up is not None:
                newv = _LOCAL._mapstack[k]
                _OPTIONS[k].set_up(newv, oldv)

    def __call__(self, fn):
        if not callable(fn): # pragma: no cover
            raise ValueError("An Env instance can only be called to decorate a function.")
        @functools.wraps(fn)
        def f(*args, **kwargs):
            with self:
                return fn(*args, **kwargs)
        return f

# Value retrieval ******************************************************************************* **
class _ThreadMapStackGetter(object):
    """Getter for env attribute"""
    def __init__(self, key):
        self.key = key

    def __call__(self, current_env_self):
        return _LOCAL._mapstack[self.key]

class _CurrentEnv(Singleton):
    """Namespace to access current values of buzzard's environment variable (see buzz.Env)

    Example
    -------
    >>> buzz.env.significant
    8.0

    """
    pass

for k in _OPTIONS.keys():
    setattr(_CurrentEnv, k, property(_ThreadMapStackGetter(k)))

env = _CurrentEnv() # pylint: disable=invalid-name



from typing import (
    Set, Dict, Tuple,
)
import uuid # For mypy

import numpy as np
import sortedcontainers

from buzzard._footprint import Footprint # For mypy
from buzzard._actors.priorities import Priorities
from buzzard._actors.message import Msg

class ActorGlobalPrioritiesWatcher(object):
    """Actor that takes care of memorizing priority information between all sub-tasks in all
    ongoing queries. Everytime a priority changes all `ActorPoolWaitingRoom` are notified.

    This class does not implement a `die` method, since destroying this actor is the same event
    as stopping the scheduler's loop. See top-level-actor.
    """

    def __init__(self):
        self._alive = True
        self._pulled_count_per_query = {}
        self.db_version = 0

        self._sorted_prod_tiles_per_cache_tile = {
        } # type: Dict[Tuple[uuid.UUID, Footprint], sortedcontainers.SortedListWithKey]
        self._pulled_count_per_query = {} # type: Dict[CachedQueryInfos, int]
        self._cache_fp_per_query = {} # type: Dict[CachedQueryInfos, Set[Footprint]]
        self.address = '/Global/GlobalPrioritiesWatcher'

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_a_query_need_those_cache_tiles(self, raster_uid, qi, cache_fps):
        """Receive message: A query started its optional collection phase and require those cache
        tiles.

        Priorities for those cache tiles should be updated if necessary.
        """
        msgs = []

        # Data structures shortcuts ********************************************
        ds0 = self._sorted_prod_tiles_per_cache_tile
        ds2 = self._cache_fp_per_query

        # Checks ***************************************************************
        assert qi not in ds2, 'received two collection phases for that query, should be 1 or 0'

        # Insert in `ds2` ******************************************************
        ds2[qi] = cache_fps

        # Insert in `ds0` and check for prio updates ***************************
        # The merger and writer are interested in this priority update
        cache_tile_updates = set()
        for cache_fp in cache_fps:
            cache_tile_key = (raster_uid, cache_fp)
            prod_tile_key = (qi, qi.dict_of_min_prod_idx_per_cache_fp[cache_fp])
            if cache_tile_key not in ds0:
                ds0[cache_tile_key] = (
                    sortedcontainers.SortedListWithKey(
                        [prod_tile_key],
                        key=lambda k: self.prio_of_prod_tile(*k)
                    )
                )
                cache_tile_updates.add(cache_tile_key)
            else:
                prev_prio = self.prio_of_prod_tile(*ds0[cache_tile_key][0])
                ds0[cache_tile_key].add(prod_tile_key)
                new_prio = self.prio_of_prod_tile(*ds0[cache_tile_key][0])
                if prev_prio != new_prio:
                    cache_tile_updates.add(cache_tile_key)

        # Emmit messages *******************************************************
        if len(cache_tile_updates) != 0:
            self.db_version += 1
            new_prio = Priorities(self, self.db_version)
            msgs += [Msg(
                '/Pool*/WaitingRoom', 'global_priorities_update',
                new_prio, frozenset(), frozenset(cache_tile_updates)
            )]

        return msgs

    def receive_cancel_this_query(self, raster_uid, qi):
        """Receive message: A query was cancelled, updated

        If this query started a collection phase, those cache tiles should be updated if necessary.
        """
        cache_tile_updates = self._get_rid_of_query(raster_uid, qi)
        if cache_tile_updates:
            self.db_version += 1
            new_prio = Priorities(self, self.db_version)
            return [Msg(
                '/Pool*/WaitingRoom', 'global_priorities_update',
                new_prio, frozenset(), frozenset(cache_tile_updates)
            )]
        else:
            return []

    def _assert_ds0_not_corrupted(self):
        ds0 = self._sorted_prod_tiles_per_cache_tile

        for cache_tile_key, sorted_prod_tiles in ds0.items():
            for ks, vs in zip(sorted_prod_tiles._keys, sorted_prod_tiles._lists):
                for k0, v in zip(ks, vs):
                    k1 = sorted_prod_tiles.key(v)
                    assert k0 == k1, """
                    cache_tile_key: {}
                    _keys:{}
                    _lists:{}
                    v:{}
                    k0:{}, k1:{}
                    """.format(
                        cache_tile_key,
                        sorted_prod_tiles._keys,
                        sorted_prod_tiles._lists,
                        v, k0, k1,
                    )

    def receive_output_queue_update(self, raster_uid, qi, produced_count, queue_size):
        """Receive message: The output queue of a query changed in size.

        If the number of arrays pulled from this queue changed:
          - Update the query priority
          - If this query started a collection phase, those cache tiles should be updated if
            necessary.
        """
        query_updates = set()
        cache_tile_updates = set()

        # Data structures shortcuts ********************************************
        ds0 = self._sorted_prod_tiles_per_cache_tile
        ds1 = self._pulled_count_per_query
        ds2 = self._cache_fp_per_query

        self._assert_ds0_not_corrupted()

        if produced_count == qi.produce_count:
            # Query is done ****************************************************
            cache_tile_updates |= self._get_rid_of_query(raster_uid, qi)

        else:
            # Query is ongoing *************************************************
            pulled_count = produced_count - queue_size
            assert pulled_count >= 0

            if qi not in ds1:
                old_pulled_count = 0
            else:
                old_pulled_count = ds1[qi]

            if old_pulled_count != pulled_count:
                query_updates.add(qi)

                # If a `new_collection_phase` was received for that query
                if qi not in ds2:
                    ds1[qi] = pulled_count
                else:
                    updates = []
                    for cache_fp in ds2[qi]:
                        min_prod_idx = qi.dict_of_min_prod_idx_per_cache_fp[cache_fp]
                        cache_tile_key = (raster_uid, cache_fp)
                        prod_tile_key = (qi, min_prod_idx)

                        prev_min_prio = self.prio_of_prod_tile(*ds0[cache_tile_key][0])
                        ds0[cache_tile_key].remove(prod_tile_key)
                        updates.append((
                            cache_tile_key, prod_tile_key, prev_min_prio
                        ))

                    self._assert_ds0_not_corrupted()
                    ds1[qi] = pulled_count
                    self._assert_ds0_not_corrupted()
                    for cache_tile_key, prod_tile_key, prev_min_prio in updates:
                        ds0[cache_tile_key].add(prod_tile_key)
                        new_min_prio = self.prio_of_prod_tile(*ds0[cache_tile_key][0])
                        if prev_min_prio != new_min_prio:
                            cache_tile_updates.add(cache_tile_key)


        if query_updates or cache_tile_updates:
            self.db_version += 1
            new_prio = Priorities(self, self.db_version)
            return [Msg(
                '/Pool*/WaitingRoom', 'global_priorities_update',
                new_prio, frozenset(query_updates), frozenset(cache_tile_updates)
            )]
        else:
            return []

    # ******************************************************************************************* **
    def prio_of_prod_tile(self, qi, prod_idx):
        # Data structures shortcuts
        ds1 = self._pulled_count_per_query

        if qi not in ds1:
            query_pulled_count = 0
        else:
            query_pulled_count = ds1[qi]

        # Priority on `produced arrays` needed soon
        prio = prod_idx - query_pulled_count
        # TODO: What if prio is negative? Is it a problem?
        return (prio,)

    def prio_of_cache_tile(self, raster_uid, cache_fp):
        # Data structures shortcuts
        ds0 = self._sorted_prod_tiles_per_cache_tile

        cache_tile_key = (raster_uid, cache_fp)
        assert cache_tile_key in ds0, (
            'Needing a priority for a cache_fp implies having it registed '
            'in this class since a qicc object was created.'
        )
        prod_tile_key = ds0[cache_tile_key][0]
        prio = self.prio_of_prod_tile(*prod_tile_key)
        return prio

    def _get_rid_of_query(self, raster_uid, qi):
        cache_tile_updates = set()

        # Data structures shortcuts
        ds0 = self._sorted_prod_tiles_per_cache_tile
        ds1 = self._pulled_count_per_query
        ds2 = self._cache_fp_per_query

        # If a `new_collection_phase` was received for that query
        # `ds0` depends of `ds1`, deleting in `ds1` before deleting in `ds0` would corrupt `ds0`
        if qi in ds2:
            for cache_fp in ds2[qi]:
                cache_tile_key = (raster_uid, cache_fp)
                prod_tile_key = (qi, qi.dict_of_min_prod_idx_per_cache_fp[cache_fp])
                prev_min_prio = self.prio_of_prod_tile(*ds0[cache_tile_key][0])
                ds0[cache_tile_key].remove(prod_tile_key)
                if len(ds0[cache_tile_key]) == 0:
                    del ds0[cache_tile_key]
                    cache_tile_updates.add(cache_tile_key)
                else:
                    new_min_prio = self.prio_of_prod_tile(*ds0[cache_tile_key][0])
                    if prev_min_prio != new_min_prio:
                        cache_tile_updates.add(cache_tile_key)
            del ds2[qi]

        # If the output queue was pulled at least once
        if qi in ds1:
            del ds1[qi]

        return cache_tile_updates

    # ******************************************************************************************* **



import collections

from buzzard._actors.message import Msg
from buzzard._actors.pool_waiting_room import ActorPoolWaitingRoom
from buzzard._actors.pool_working_room import ActorPoolWorkingRoom
from buzzard._actors.global_priorities_watcher import ActorGlobalPrioritiesWatcher

class ActorTopLevel(object):
    """Actor that takes care of the lifetime of all other actors.

    This is the only actor that is instantiated by the scheduler. All other actors are
    instantiated from here.

    This class does not implement a `die` method, since destroying this actor is the same event
    as stopping the scheduler's loop. If a destruction is ever needed, call a die method from
    the scheduler using the `top_level_actor` variable.
    """
    def __init__(self):
        self._rasters = set()
        self._rasters_per_pool = collections.defaultdict(list)

        self._actor_addresses_of_raster = {}
        self._actor_addresses_of_pool = {}

        self._primed = False
        self._alive = True

    address = '/Global/TopLevel'

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def ext_receive_prime(self):
        """Receive message sent by something else than an actor, still treated synchronously: Prime
        yourself, we are about to start the show!
        """
        assert not self._primed
        self._primed = True
        return [
            ActorGlobalPrioritiesWatcher()
        ]

    def ext_receive_new_raster(self, raster):
        """Receive message sent by something else than an actor, still treated synchronously: There
        is a new raster

        Parameter
        ---------
        raster: _a_recipe_raster.ABackRecipeRaster
        """
        msgs = []
        self._rasters.add(raster)
        raster.debug_mngr.event('raster_started', raster.facade_proxy)

        # Instantiate raster's actors ******************************************
        actors = raster.create_actors()
        msgs += actors
        self._actor_addresses_of_raster[raster] = [
            actor.address
            for actor in actors
        ]
        del actors

        # Instantiate pools' actors ********************************************
        pools = {
            id(pool): pool
            for attr in [
                'computation_pool', 'merge_pool', 'io_pool', 'resample_pool',
            ]
            if hasattr(raster, attr)
            for pool in [getattr(raster, attr)]
            if pool is not None
        }
        for pool_id, pool in pools.items():
            if pool_id not in self._rasters_per_pool:
                actors = [
                    ActorPoolWaitingRoom(pool),
                    ActorPoolWorkingRoom(pool),
                ]
                msgs += actors

                self._actor_addresses_of_pool[pool_id] = [
                    actor.address
                    for actor in actors
                ]

            self._rasters_per_pool[pool_id].append(raster)

        return msgs

    def ext_receive_kill_raster(self, raster):
        """Receive message sent by something else than an actor, still treated synchronously: An
        actor is closing

        Parameter
        ---------
        raster: _a_recipe_raster.ABackRecipeRaster
        """
        msgs = []
        self._rasters.remove(raster)
        raster.debug_mngr.event('raster_stopped', raster.facade_proxy)

        # Deleting raster's actors *********************************************
        # Deal with QueriesHandler first.
        # TODO: Should the order of 'die' messages be enforced somewhere else?
        msgs += [
            Msg(address, 'die')
            for address in sorted(
                    self._actor_addresses_of_raster[raster],
                    key=lambda address: 'QueriesHandler' not in address,
            )
        ]
        del self._actor_addresses_of_raster[raster]

        # Deleting pools' actors ***********************************************
        pools = {
            id(pool): pool
            for attr in [
                'computation_pool', 'merge_pool', 'io_pool', 'resample_pool',
            ]
            if hasattr(raster, attr)
            for pool in [getattr(raster, attr)]
            if pool is not None
        }
        for pool_id in pools.keys():
            self._rasters_per_pool[pool_id].remove(raster)
            if len(self._rasters_per_pool[pool_id]) == 0:
                del self._rasters_per_pool[pool_id]
                msgs += [
                    Msg(actor_adress, 'die')
                    for actor_adress in self._actor_addresses_of_pool[pool_id]
                ]
                del self._actor_addresses_of_pool[pool_id]

        return msgs

    # ******************************************************************************************* **
